"""
PC-BASIC - modes.py
Emulated video modes

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct

try:
    import numpy
except ImportError:
    numpy = None

from ...compat import xrange, int2byte

from .. import values
from ..base import error


# SCREEN 10 EGA pseudocolours, blink state 0 and 1
INTENSITY_EGA_MONO_0 = (0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff)
INTENSITY_EGA_MONO_1 = (0x00, 0xaa, 0xff, 0x00, 0xaa, 0xff, 0x00, 0xaa, 0xff)
# MDA text intensities: black, dark green, green, bright green
INTENSITY_MDA_MONO = (0x00, 0x40, 0xc0, 0xff)

# default 16-color and ega palettes
CGA16_PALETTE = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15)
EGA_PALETTE = (0, 1, 2, 3, 4, 5, 20, 7, 56, 57, 58, 59, 60, 61, 62, 63)
EGA_MONO_PALETTE = (0, 4, 1, 8)
# adding dark-green (foreground for some exceptional attributes) as #3
MDA_PALETTE = (0, 2, 3, 1)
CGA2_PALETTE = (0, 15)

# http://qbhlp.uebergeord.net/screen-statement-details-colors.html
# underline/intensity/reverse video attributes are slightly different from mda
# attributes 1, 9 should have underlining.
# EGA_MONO_TEXT_PALETTE = (0, 1, 1, 1, 1, 1, 1, 1, 0, 2, 2, 2, 2, 2, 2, 0)
# MDA_PALETTE = (0, 1, 1, 1, 1, 1, 1, 1, 0, 2, 2, 2, 2, 2, 2, 2)

# CGA mono intensities
INTENSITY16 = range(0x00, 0x100, 0x11)
# CGA colours
COLOURS16 = (
    (0x00, 0x00, 0x00), (0x00, 0x00, 0xaa), (0x00, 0xaa, 0x00), (0x00, 0xaa, 0xaa),
    (0xaa, 0x00, 0x00), (0xaa, 0x00, 0xaa), (0xaa, 0x55, 0x00), (0xaa, 0xaa, 0xaa),
    (0x55, 0x55, 0x55), (0x55, 0x55, 0xff), (0x55, 0xff, 0x55), (0x55, 0xff, 0xff),
    (0xff, 0x55, 0x55), (0xff, 0x55, 0xff), (0xff, 0xff, 0x55), (0xff, 0xff, 0xff)
)
# EGA colours
COLOURS64 = (
    (0x00, 0x00, 0x00), (0x00, 0x00, 0xaa), (0x00, 0xaa, 0x00), (0x00, 0xaa, 0xaa),
    (0xaa, 0x00, 0x00), (0xaa, 0x00, 0xaa), (0xaa, 0xaa, 0x00), (0xaa, 0xaa, 0xaa),
    (0x00, 0x00, 0x55), (0x00, 0x00, 0xff), (0x00, 0xaa, 0x55), (0x00, 0xaa, 0xff),
    (0xaa, 0x00, 0x55), (0xaa, 0x00, 0xff), (0xaa, 0xaa, 0x55), (0xaa, 0xaa, 0xff),
    (0x00, 0x55, 0x00), (0x00, 0x55, 0xaa), (0x00, 0xff, 0x00), (0x00, 0xff, 0xaa),
    (0xaa, 0x55, 0x00), (0xaa, 0x55, 0xaa), (0xaa, 0xff, 0x00), (0xaa, 0xff, 0xaa),
    (0x00, 0x55, 0x55), (0x00, 0x55, 0xff), (0x00, 0xff, 0x55), (0x00, 0xff, 0xff),
    (0xaa, 0x55, 0x55), (0xaa, 0x55, 0xff), (0xaa, 0xff, 0x55), (0xaa, 0xff, 0xff),
    (0x55, 0x00, 0x00), (0x55, 0x00, 0xaa), (0x55, 0xaa, 0x00), (0x55, 0xaa, 0xaa),
    (0xff, 0x00, 0x00), (0xff, 0x00, 0xaa), (0xff, 0xaa, 0x00), (0xff, 0xaa, 0xaa),
    (0x55, 0x00, 0x55), (0x55, 0x00, 0xff), (0x55, 0xaa, 0x55), (0x55, 0xaa, 0xff),
    (0xff, 0x00, 0x55), (0xff, 0x00, 0xff), (0xff, 0xaa, 0x55), (0xff, 0xaa, 0xff),
    (0x55, 0x55, 0x00), (0x55, 0x55, 0xaa), (0x55, 0xff, 0x00), (0x55, 0xff, 0xaa),
    (0xff, 0x55, 0x00), (0xff, 0x55, 0xaa), (0xff, 0xff, 0x00), (0xff, 0xff, 0xaa),
    (0x55, 0x55, 0x55), (0x55, 0x55, 0xff), (0x55, 0xff, 0x55), (0x55, 0xff, 0xff),
    (0xff, 0x55, 0x55), (0xff, 0x55, 0xff), (0xff, 0xff, 0x55), (0xff, 0xff, 0xff)
)

# composite palettes, see http://nerdlypleasures.blogspot.co.uk/2013_11_01_archive.html
COMPOSITE = {
    'cga_old': [
        (0x00, 0x00, 0x00),   (0x00, 0x71, 0x00),   (0x00, 0x3f, 0xff),   (0x00, 0xab, 0xff),
        (0xc3, 0x00, 0x67),   (0x73, 0x73, 0x73),   (0xe6, 0x39, 0xff),   (0x8c, 0xa8, 0xff),
        (0x53, 0x44, 0x00),   (0x00, 0xcd, 0x00),   (0x73, 0x73, 0x73),   (0x00, 0xfc, 0x7e),
        (0xff, 0x39, 0x00),   (0xe2, 0xca, 0x00),   (0xff, 0x7c, 0xf4),   (0xff, 0xff, 0xff)
    ],
    'cga': [
        (0x00, 0x00, 0x00),   (0x00, 0x6a, 0x2c),   (0x00, 0x39, 0xff),   (0x00, 0x94, 0xff),
        (0xca, 0x00, 0x2c),   (0x77, 0x77, 0x77),   (0xff, 0x31, 0xff),   (0xc0, 0x98, 0xff),
        (0x1a, 0x57, 0x00),   (0x00, 0xd6, 0x00),   (0x77, 0x77, 0x77),   (0x00, 0xf4, 0xb8),
        (0xff, 0x57, 0x00),   (0xb0, 0xdd, 0x00),   (0xff, 0x7c, 0xb8),   (0xff, 0xff, 0xff)
    ],
    'tandy': [
        (0x00, 0x00, 0x00),   (0x7c, 0x30, 0x00),   (0x00, 0x75, 0x00),   (0x00, 0xbe, 0x00),
        (0x00, 0x47, 0xee),   (0x77, 0x77, 0x77),   (0x00, 0xbb, 0xc4),   (0x00, 0xfb, 0x3f),
        (0xb2, 0x0f, 0x9d),   (0xff, 0x1e, 0x0f),   (0x77, 0x77, 0x77),   (0xff, 0xb8, 0x00),
        (0xb2, 0x44, 0xff),   (0xff, 0x78, 0xff),   (0x4b, 0xba, 0xff),   (0xff, 0xff, 0xff)
    ],
    'pcjr': [
        (0x00, 0x00, 0x00),
        (0x98, 0x20, 0xcb),   (0x9f, 0x1c, 0x00),   (0xff, 0x11, 0x71),   (0x00, 0x76, 0x00),
        (0x77, 0x77, 0x77),   (0x5b, 0xaa, 0x00),   (0xff, 0xa5, 0x00),   (0x00, 0x4e, 0xcb),
        (0x74, 0x53, 0xff),   (0x77, 0x77, 0x77),   (0xff, 0x79, 0xff),   (0x00, 0xc8, 0x71),
        (0x00, 0xcc, 0xff),   (0x00, 0xfa, 0x00),   (0xff, 0xff, 0xff)
    ]
}

MONO_TINT = {
    'green': [0, 255, 0],
    'amber': [255, 128, 0],
    'grey': [255, 255, 255],
    'mono': [0, 255, 0],
}


###############################################################################
# Low level video (mainly about colours; mode object factory)

class Video(object):
    """Low-level display operations."""

    def __init__(self, capabilities, monitor, low_intensity, aspect, video_mem_size):
        """Initialise colour sets."""
        # public members - used by VideoMode
        # video adapter type - cga, ega, etc
        if capabilities == 'ega' and monitor in MONO_TINT:
            capabilities = 'ega_mono'
        self.capabilities = capabilities
        # monochrome tint in rgb
        self.mono_tint = MONO_TINT.get(monitor, MONO_TINT['green'])
        # emulated monitor type - rgb, composite, mono
        self.monitor = 'mono' if monitor in MONO_TINT else monitor
        # build 16-greyscale and 16-colour sets
        self.colours16_mono = tuple(
            tuple(tint*i//255 for tint in self.mono_tint) for i in INTENSITY16
        )
        # NTSC colorburst settings
        if monitor == 'mono':
            self.colours16 = list(self.colours16_mono)
        else:
            self.colours16 = list(COLOURS16)
        # CGA 4-colour palette / mode 5 settings
        # palette 1: Black, Ugh, Yuck, Bleah, choice of low & high intensity
        # palette 0: Black, Green, Red, Brown/Yellow, low & high intensity
        # tandy/pcjr have high-intensity white, but low-intensity colours
        # mode 5 (SCREEN 1 + colorburst on RGB) has red instead of magenta
        if capabilities in ('pcjr', 'tandy'):
            # pcjr does not have mode 5
            self.cga4_palettes = {0: (0, 2, 4, 6), 1: (0, 3, 5, 15), 5: None}
        elif low_intensity:
            self.cga4_palettes = {0: (0, 2, 4, 6), 1: (0, 3, 5, 7), 5: (0, 3, 4, 7)}
        else:
            self.cga4_palettes = {0: (0, 10, 12, 14), 1: (0, 11, 13, 15), 5: (0, 11, 12, 15)}
        self.cga4_palette = list(self.cga4_palettes[1])
        self.cga4_palette_num = 1
        self.cga_mode_5 = False
        # screen aspect ratio, for CIRCLE
        self.aspect = aspect
        # set up text_data and mode_data
        self.prepare_modes(video_mem_size)

    def get_mode(self, number, width=None):
        """Retrieve graphical mode by screen number."""
        try:
            if number:
                return self._mode_data[number]
            else:
                return self._text_data[width]
        except KeyError:
            # no such mode
            raise error.BASICError(error.IFC)

    def toggle_colour(self, has_colour):
        """Toggle between colour and monochrome (for NTSC colorburst)."""
        # note that colours16 member is only used in certain mode/adapter combinations
        # e.g. in text mode it's only used for 'cga', 'cga_old', 'pcjr', 'tandy'
        if has_colour:
            self.colours16[:] = COLOURS16
        else:
            self.colours16[:] = self.colours16_mono

    def set_cga4_palette(self, num):
        """set the default 4-colour CGA palette."""
        self.cga4_palette_num = num
        # we need to copy into cga4_palette as it's referenced by mode.palette
        if self.cga_mode_5 and self.capabilities in ('cga', 'cga_old'):
            self.cga4_palette[:] = self.cga4_palettes[5]
        else:
            self.cga4_palette[:] = self.cga4_palettes[num]

    def set_colorburst(self, on, is_cga):
        """Set the NTSC colorburst bit."""
        # On a composite monitor with CGA adapter (not EGA, VGA):
        # - on SCREEN 2 this enables artifacting
        # - on SCREEN 1 and 0 this switches between colour and greyscale
        # On an RGB monitor:
        # - on SCREEN 1 this switches between mode 4/5 palettes (RGB)
        # - ignored on other screens
        colorburst_capable = self.capabilities in ('cga', 'cga_old', 'tandy', 'pcjr')
        if is_cga and self.monitor != 'composite':
            # ega ignores colorburst; tandy and pcjr have no mode 5
            self.cga_mode_5 = not on
            self.set_cga4_palette(1)
        else:
            self.toggle_colour(self.monitor != 'mono' and (on or self.monitor != 'composite'))
        return on and colorburst_capable

    def get_allowed_widths(self):
        """Get allowed screen widths."""
        return set(
            mode.width for mode in list(self._text_data.values()) + list(self._mode_data.values())
        )

    ###########################################################################
    # video modes

    def prepare_modes(self, video_mem_size):
        """Build lists of allowed graphics modes."""
        video_mem_size = int(video_mem_size)
        # initialise tinted monochrome palettes
        colours_ega_mono_0 = tuple(
            tuple(tint*i//255 for tint in self.mono_tint) for i in INTENSITY_EGA_MONO_0
        )
        colours_ega_mono_1 = tuple(
            tuple(tint*i//255 for tint in self.mono_tint) for i in INTENSITY_EGA_MONO_1
        )
        colours_mda_mono = tuple(
            tuple(tint*i//255 for tint in self.mono_tint) for i in INTENSITY_MDA_MONO
        )
        # Tandy/PCjr pixel aspect ratio is different from normal
        # suggesting screen aspect ratio is not 4/3.
        # Tandy pixel aspect ratios, experimentally found with CIRCLE:
        # screen 2, 6:     48/100   normal if aspect = 3072, 2000
        # screen 1, 4, 5:  96/100   normal if aspect = 3072, 2000
        # screen 3:      1968/1000
        # screen 3 is strange, slighly off the 192/100 you'd expect
        graphics_mode = {
            # 04h 320x200x4  16384B 2bpp 0xb8000    screen 1
            # tandy:2 pages if 32k memory; ega: 1 page only
            '320x200x4': CGAMode(
                '320x200x4', 320, 200, 25, 40, 3,
                self.cga4_palette, self.colours16, bitsperpixel=2,
                interleave_times=2, bank_size=0x2000,
                aspect=self.aspect,
                num_pages=(
                    video_mem_size // (2*0x2000)
                    if self.capabilities in ('pcjr', 'tandy')
                    else 1
                )
            ),
            # 06h 640x200x2  16384B 1bpp 0xb8000    screen 2
            '640x200x2': CGAMode(
                '640x200x2', 640, 200, 25, 80, 1,
                CGA2_PALETTE, self.colours16, bitsperpixel=1,
                interleave_times=2, bank_size=0x2000, num_pages=1,
                aspect=self.aspect,
                supports_artifacts=True
            ),
            # 08h 160x200x16 16384B 4bpp 0xb8000    PCjr/Tandy screen 3
            '160x200x16': CGAMode(
                '160x200x16', 160, 200, 25, 20, 15,
                CGA16_PALETTE, self.colours16, bitsperpixel=4,
                interleave_times=2, bank_size=0x2000,
                num_pages=video_mem_size//(2*0x2000),
                pixel_aspect=(1968, 1000), cursor_index=3
            ),
            #     320x200x4  16384B 2bpp 0xb8000   Tandy/PCjr screen 4
            '320x200x4pcjr': CGAMode(
                '320x200x4pcjr', 320, 200, 25, 40, 3,
                self.cga4_palette, self.colours16, bitsperpixel=2,
                interleave_times=2, bank_size=0x2000,
                num_pages=video_mem_size//(2*0x2000),
                aspect=self.aspect,
                cursor_index=3
            ),
            # 09h 320x200x16 32768B 4bpp 0xb8000    Tandy/PCjr screen 5
            '320x200x16pcjr': CGAMode(
                '320x200x16pcjr', 320, 200, 25, 40, 15,
                CGA16_PALETTE, self.colours16, bitsperpixel=4,
                interleave_times=4, bank_size=0x2000,
                num_pages=video_mem_size//(4*0x2000),
                aspect=self.aspect,
                cursor_index=3
            ),
            # 0Ah 640x200x4  32768B 2bpp 0xb8000   Tandy/PCjr screen 6
            '640x200x4': Tandy6Mode(
                '640x200x4', 640, 200, 25, 80, 3,
                self.cga4_palette, self.colours16, bitsperpixel=2,
                interleave_times=4, bank_size=0x2000,
                num_pages=video_mem_size//(4*0x2000),
                aspect=self.aspect,
                cursor_index=3
            ),
            # 0Dh 320x200x16 32768B 4bpp 0xa0000    EGA screen 7
            '320x200x16': EGAMode(
                '320x200x16', 320, 200, 25, 40, 15,
                CGA16_PALETTE, self.colours16, bitsperpixel=4,
                num_pages=video_mem_size//(4*0x2000),
                aspect=self.aspect,
                interleave_times=1, bank_size=0x2000
            ),
            # 0Eh 640x200x16    EGA screen 8
            '640x200x16': EGAMode(
                '640x200x16', 640, 200, 25, 80, 15,
                CGA16_PALETTE, self.colours16, bitsperpixel=4,
                num_pages=video_mem_size//(4*0x4000),
                aspect=self.aspect,
                interleave_times=1, bank_size=0x4000
            ),
            # 10h 640x350x16    EGA screen 9
            '640x350x16': EGAMode(
                '640x350x16', 640, 350, 25, 80, 15,
                EGA_PALETTE, COLOURS64, bitsperpixel=4,
                num_pages=video_mem_size//(4*0x8000),
                aspect=self.aspect,
                interleave_times=1, bank_size=0x8000
            ),
            # 0Fh 640x350x4     EGA monochrome screen 10
            '640x350x4': EGAMode(
                '640x350x16', 640, 350, 25, 80, 1,
                EGA_MONO_PALETTE, colours_ega_mono_0, bitsperpixel=2,
                interleave_times=1, bank_size=0x8000,
                num_pages=video_mem_size//(2*0x8000),
                aspect=self.aspect,
                colours1=colours_ega_mono_1, has_blink=True,
                planes_used=(1, 3)
            ),
            # 40h 640x400x2   1bpp  olivetti screen 3
            '640x400x2': CGAMode(
                '640x400x2', 640, 400, 25, 80, 1,
                CGA2_PALETTE, self.colours16, bitsperpixel=1,
                interleave_times=4, bank_size=0x2000,
                num_pages=1,
                aspect=self.aspect,
                has_blink=True
            ),
            # hercules screen 3
            '720x348x2': CGAMode(
                # this actually produces 350, not 348
                '720x348x2', 720, 350, 25, 80, 1,
                CGA2_PALETTE, self.colours16_mono, bitsperpixel=1,
                interleave_times=4, bank_size=0x2000,
                num_pages=2,
                aspect=self.aspect,
                has_blink=True
            ),
        }
        if self.capabilities == 'vga':
            # technically, VGA text does have underline
            # but it's set to an invisible scanline
            # so not, so long as we're not allowing to set the scanline
            self._text_data = {
                40: TextMode('vgatext40', 25, 40, 16, 9, 7, EGA_PALETTE, COLOURS64, num_pages=8),
                80: TextMode('vgatext80', 25, 80, 16, 9, 7, EGA_PALETTE, COLOURS64, num_pages=4)
            }
            self._mode_data = {
                1: graphics_mode['320x200x4'],
                2: graphics_mode['640x200x2'],
                7: graphics_mode['320x200x16'],
                8: graphics_mode['640x200x16'],
                9: graphics_mode['640x350x16']
            }
        elif self.capabilities == 'ega':
            self._text_data = {
                40: TextMode('egatext40', 25, 40, 14, 8, 7, EGA_PALETTE, COLOURS64, num_pages=8),
                80: TextMode('egatext80', 25, 80, 14, 8, 7, EGA_PALETTE, COLOURS64, num_pages=4)
            }
            self._mode_data = {
                1: graphics_mode['320x200x4'],
                2: graphics_mode['640x200x2'],
                7: graphics_mode['320x200x16'],
                8: graphics_mode['640x200x16'],
                9: graphics_mode['640x350x16']
            }
        elif self.capabilities == 'ega_mono':
            self._text_data = {
                40: MonoTextMode(
                    'ega_monotext40', 25, 40, 14, 8, 7,
                    MDA_PALETTE, colours_mda_mono, is_mono=True, num_pages=8
                ),
                80: MonoTextMode(
                    'ega_monotext80', 25, 80, 14, 8, 7,
                    MDA_PALETTE, colours_mda_mono, is_mono=True, num_pages=4
                )
            }
            self._mode_data = {
                10: graphics_mode['640x350x4']
            }
        elif self.capabilities == 'mda':
            self._text_data = {
                40: MonoTextMode(
                    'mdatext40', 25, 40, 14, 9, 7,
                    MDA_PALETTE, colours_mda_mono, is_mono=True, num_pages=1
                ),
                80: MonoTextMode(
                    'mdatext80', 25, 80, 14, 9, 7,
                    MDA_PALETTE, colours_mda_mono, is_mono=True, num_pages=1
                )
            }
            self._mode_data = {}
        elif self.capabilities in ('cga', 'cga_old', 'pcjr', 'tandy'):
            if self.capabilities == 'tandy':
                self._text_data = {
                    40: TextMode(
                        'tandytext40', 25, 40, 9, 8, 7, CGA16_PALETTE, self.colours16, num_pages=8
                    ),
                    80: TextMode(
                        'tandytext80', 25, 80, 9, 8, 7, CGA16_PALETTE, self.colours16, num_pages=4
                    )
                }
            else:
                self._text_data = {
                    40: TextMode(
                        'cgatext40', 25, 40, 8, 8, 7,
                        CGA16_PALETTE, self.colours16, num_pages=8
                    ),
                    80: TextMode(
                        'cgatext80', 25, 80, 8, 8, 7,
                        CGA16_PALETTE, self.colours16, num_pages=4
                    )
                }
            if self.capabilities in ('cga', 'cga_old'):
                self._mode_data = {
                    1: graphics_mode['320x200x4'],
                    2: graphics_mode['640x200x2']
                }
            else:
                self._mode_data = {
                    1: graphics_mode['320x200x4'],
                    2: graphics_mode['640x200x2'],
                    3: graphics_mode['160x200x16'],
                    4: graphics_mode['320x200x4pcjr'],
                    5: graphics_mode['320x200x16pcjr'],
                    6: graphics_mode['640x200x4']
                }
        elif self.capabilities == 'hercules':
            # herc attributes shld distinguish black, dim, normal, bright
            # see http://www.seasip.info/VintagePC/hercplus.html
            self._text_data = {
                40: MonoTextMode(
                    'herculestext40', 25, 40, 14, 9, 7,
                    MDA_PALETTE, colours_mda_mono, is_mono=True, num_pages=2
                ),
                80: MonoTextMode(
                    'herculestext80', 25, 80, 14, 9, 7,
                    MDA_PALETTE, colours_mda_mono, is_mono=True, num_pages=2
                )
            }
            self._mode_data = {
                3: graphics_mode['720x348x2']
            }
        elif self.capabilities == 'olivetti':
            self._text_data = {
                40: TextMode(
                    'olivettitext40', 25, 40, 16, 8, 7, CGA16_PALETTE, self.colours16, num_pages=8
                ),
                80: TextMode(
                    'olivettitext80', 25, 80, 16, 8, 7, CGA16_PALETTE, self.colours16, num_pages=4
                )
            }
            self._mode_data = {
                1: graphics_mode['320x200x4'],
                2: graphics_mode['640x200x2'],
                3: graphics_mode['640x400x2']
            }
            # on Olivetti M24, all numbers 3-255 give the same altissima risoluzione
            for mode in range(4, 256):
                self._mode_data[mode] = graphics_mode['640x400x2']


class VideoMode(object):
    """Base class for video modes."""
    def __init__(
                self, name, height, width,
                font_height, font_width,
                attr, palette, colours,
                num_pages, has_blink,
                video_segment, page_size
            ):
        """Initialise video mode settings."""
        self.is_text_mode = False
        self.name = name
        self.height = int(height)
        self.width = int(width)
        self.font_height = int(font_height)
        self.font_width = int(font_width)
        self.pixel_height = self.height*self.font_height
        self.pixel_width = self.width*self.font_width
        self.attr = int(attr)
        # palette is a reference (changes with cga_mode_5 and cga4_palette_num)
        self.palette = palette
        self.num_attr = len(palette)
        # colours is a reference (changes with colorburst on composite)
        self.colours = colours
        # colours1 is only used by EGA mono mode
        self.colours1 = None
        # this is used by interface plugins (which use this object as mode_info)
        self.has_blink = has_blink
        self.video_segment = int(video_segment)
        self.page_size = int(page_size)
        self.num_pages = int(num_pages) # or video_mem_size // self.page_size)

    def split_attr(self, attr):
        """Split attribute byte into constituent parts."""
        # 7  6 5 4  3 2 1 0
        # Bl b b b  f f f f
        back = (attr >> 4) & 7
        blink = (attr >> 7) == 1
        fore = attr & 0xf
        underline = False
        return fore, back, blink, underline

    def pixel_to_text_pos(self, x, y):
        """Convert pixel position to text position."""
        return 1 + y // self.font_height, 1 + x // self.font_width

    def pixel_to_text_area(self, x0, y0, x1, y1):
        """Convert from pixel area to text area."""
        col0 = min(self.width, max(1, 1 + x0 // self.font_width))
        row0 = min(self.height, max(1, 1 + y0 // self.font_height))
        col1 = min(self.width, max(1, 1 + x1 // self.font_width))
        row1 = min(self.height, max(1, 1 + y1 // self.font_height))
        return row0, col0, row1, col1

    def text_to_pixel_area(self, row0, col0, row1, col1):
        """Convert area text area to pixel area."""
        # area bounds are all inclusive
        return (
            (col0-1) * self.font_width, (row0-1) * self.font_height,
            (col1-col0+1) * self.font_width-1, (row1-row0+1) * self.font_height-1
        )

    def get_all_memory(self, screen):
        """Obtain a copy of all video memory."""
        return self.get_memory(screen, self.video_segment*0x10, self.page_size*self.num_pages)

    def set_all_memory(self, screen, mem_copy):
        """Restore a copy of all video memory."""
        return self.set_memory(screen, self.video_segment*0x10, mem_copy)

    def get_memory(self, screen, addr, num_bytes):
        """Retrieve bytes from video memory, stub."""

    def set_memory(self, screen, addr, bytes):
        """Set bytes in video memory, stub."""


class TextMode(VideoMode):
    """Default settings for a text mode."""

    def __init__(
            self, name, height, width,
            font_height, font_width, attr, palette, colours,
            num_pages, is_mono=False, has_blink=True
        ):
        """Initialise video mode settings."""
        video_segment = 0xb000 if is_mono else 0xb800
        page_size = 0x1000 if width == 80 else 0x800
        VideoMode.__init__(
            self, name, height, width,
            font_height, font_width, attr, palette, colours,
            num_pages, has_blink, video_segment, page_size
        )
        self.is_text_mode = True
        self.num_attr = 32

    def get_memory(self, screen, addr, num_bytes):
        """Retrieve bytes from textmode video memory."""
        addr -= self.video_segment*0x10
        mem_bytes = bytearray(num_bytes)
        for i in xrange(num_bytes):
            page = (addr+i) // self.page_size
            offset = (addr+i) % self.page_size
            ccol = 1 + (offset % (self.width*2)) // 2
            crow = 1 + offset // (self.width*2)
            try:
                if (addr+i) % 2:
                    mem_bytes[i] = screen.text_screen.text.get_attr(page, crow, ccol)
                else:
                    mem_bytes[i] = screen.text_screen.text.get_char(page, crow, ccol)
            except IndexError:
                pass
        return mem_bytes

    def set_memory(self, screen, addr, mem_bytes):
        """Set bytes in textmode video memory."""
        addr -= self.video_segment*0x10
        last_row = 0
        for i in xrange(len(mem_bytes)):
            page = (addr+i) // self.page_size
            offset = (addr+i) % self.page_size
            ccol = 1 + (offset % (self.width*2)) // 2
            crow = 1 + offset // (self.width*2)
            try:
                if (addr+i) % 2:
                    c = screen.text_screen.text.get_char(page, crow, ccol)
                    a = mem_bytes[i]
                else:
                    c = mem_bytes[i]
                    a = screen.text_screen.text.get_attr(page, crow, ccol)
                screen.text_screen.text.put_char_attr(page, crow, ccol, int2byte(c), a)
                if last_row > 0 and last_row != crow:
                    screen.text_screen.refresh_range(page, last_row, 1, self.width)
            except IndexError:
                pass
            last_row = crow
        if last_row >= 1 and last_row <= self.height and page >= 0 and page < self.num_pages:
            screen.text_screen.refresh_range(page, last_row, 1, self.width)


class MonoTextMode(TextMode):
    """MDA-style text mode with underlining."""
    def split_attr(self, attr):
        """Split attribute byte into constituent parts."""
        # MDA text attributes: http://www.seasip.info/VintagePC/mda.html
        # see also http://support.microsoft.com/KB/35148
        # don't try to change this with PALETTE, it won't work correctly
        underline = (attr % 8) == 1
        blink = (attr & 0x80) != 0
        # background is almost always black
        back = 0
        # intensity set by bit 3
        fore = 1 if not (attr & 0x8) else 2
        # exceptions
        if attr in (0x00, 0x08, 0x80, 0x88):
            fore, back = 0, 0
        elif attr in (0x70, 0xf0):
            fore, back = 0, 1
        elif attr in (0x78, 0xf8):
            fore, back = 3, 1
        return fore, back, blink, underline


# helper functions: convert between attribute lists and byte arrays

if numpy:
    def bytes_to_interval(byte_array, pixels_per_byte, mask=1):
        """Convert masked attributes packed into bytes to a scanline interval."""
        bpp = 8 // pixels_per_byte
        attrmask = (1<<bpp) - 1
        bitval = numpy.array([128, 64, 32, 16, 8, 4, 2, 1], dtype=numpy.uint8)
        bitmask = bitval[0::bpp]
        for i in range(1, bpp):
            bitmask |= bitval[i::bpp]
        pre_mask = numpy.tile(bitmask, len(byte_array))
        post_shift = numpy.tile(
            numpy.array([7, 6, 5, 4, 3, 2, 1, 0])[(bpp-1)::bpp], len(byte_array)
        )
        attrs = numpy.right_shift(
            numpy.repeat(numpy.array(byte_array).astype(int), pixels_per_byte) & pre_mask,
            post_shift
        ) & attrmask
        return numpy.array(attrs) * mask

else:
    def bytes_to_interval(byte_array, pixels_per_byte, mask=1):
        """Convert masked attributes packed into bytes to a scanline interval."""
        bpp = 8 // pixels_per_byte
        attrmask = (1<<bpp) - 1
        return [
            ((byte >> (8-bpp-shift)) & attrmask) * mask
            for byte in byte_array for shift in range(0, 8, bpp)
        ]

if numpy:
    def interval_to_bytes(colours, pixels_per_byte, plane=0):
        """Convert a scanline interval into masked attributes packed into bytes."""
        num_pixels = len(colours)
        num_bytes, odd_out = divmod(num_pixels, pixels_per_byte)
        if odd_out:
            num_bytes += 1
        bpp = 8 // pixels_per_byte
        attrmask = (1<<bpp) - 1
        colours = numpy.array(colours).astype(int)
        if odd_out:
            colours.resize(len(colours)+pixels_per_byte-odd_out)
        shift = numpy.tile(numpy.array([7, 6, 5, 4, 3, 2, 1, 0])[(bpp-1)::bpp], num_bytes)
        attrs = numpy.right_shift(colours, plane)
        attrs = numpy.left_shift(attrs & attrmask, shift)
        # below is much faster than:
        #   return list([ sum(attrs[i:i+pixels_per_byte])
        #                 for i in range(0, len(attrs), pixels_per_byte) ])
        # and anything involving numpy.array_split or numpy.dot is even slower.
        # numpy.roll is ok but this is the fastest I've found:
        nattrs = attrs[0::pixels_per_byte]
        for i in range(1, pixels_per_byte):
            nattrs |= attrs[i::pixels_per_byte]
        return bytearray(list(nattrs))

else:
    def interval_to_bytes(colours, pixels_per_byte, plane=0):
        """Convert a scanline interval into masked attributes packed into bytes."""
        num_pixels = len(colours)
        num_bytes, odd_out = divmod(num_pixels, pixels_per_byte)
        if odd_out:
            num_bytes += 1
        bpp = 8//pixels_per_byte
        attrmask = (1<<bpp) - 1
        colours = list(colours)
        byte_list = bytearray(num_bytes)
        shift, byte = -1, -1
        for x in xrange(num_pixels):
            if shift < 0:
                shift = 8 - bpp
                byte += 1
            byte_list[byte] |= ((colours[x] >> plane) & attrmask) << shift
            shift -= bpp
        return byte_list

def walk_memory(self, addr, num_bytes, factor=1):
    """Yield parts of graphics memory corresponding to pixels."""
    # factor supports tandy-6 mode, which has 8 pixels per 2 bytes
    # with alternating planes in even and odd bytes (i.e. ppb==8)
    ppb = factor * self.ppb
    page_size = self.page_size//factor
    bank_size = self.bank_size//factor
    row_size = self.bytes_per_row//factor
    # first row
    page, x, y = self.get_coords(addr)
    offset = min(row_size - x//ppb, num_bytes)
    if self.coord_ok(page, x, y):
        yield page, x, y, 0, offset
    # full rows
    bank_offset, page_offset, start_y = 0, 0, y
    while page_offset + bank_offset + offset < num_bytes:
        y += self.interleave_times
        # not an integer number of rows in a bank
        if offset >= bank_size:
            bank_offset += bank_size
            start_y += 1
            offset, y = 0, start_y
            if bank_offset >= page_size:
                page_offset += page_size
                page += 1
                bank_offset, offset = 0, 0
                y, start_y = 0, 0
        if self.coord_ok(page, 0, y):
            ofs = page_offset + bank_offset + offset
            if ofs + row_size > num_bytes:
                yield page, 0, y, ofs, num_bytes - ofs
            else:
                yield page, 0, y, ofs, row_size
        offset += row_size

def sprite_size_to_record_ega(self, dx, dy):
    """Write 4-byte record of sprite size in EGA modes."""
    return struct.pack('<HH', dx, dy)

def record_to_sprite_size_ega(self, byte_array):
    """Read 4-byte record of sprite size in EGA modes."""
    return struct.unpack('<HH', byte_array[0:4])

def sprite_to_array_ega(self, attrs, dx, dy, byte_array, offs):
    """Build the sprite byte array in EGA modes."""
    # for EGA modes, sprites have 8 pixels per byte
    # with colour planes in consecutive rows
    # each new row is aligned on a new byte
    #
    # this is much faster for wide selections
    # but for narrow selections storing in an array and indexing take longer
    # than just getting each pixel separately
    row_bytes = (dx+7) // 8
    length = dy * self.bitsperpixel * row_bytes
    if offs+length > len(byte_array):
        raise ValueError('Sprite exceeds array byte size')
    byte_array[offs:offs+length] = b'\0'*length
    for row in attrs:
        for plane in range(self.bitsperpixel):
            byte_array[offs:offs+row_bytes] = interval_to_bytes(row, 8, plane)
            offs += row_bytes

# elementwise OR, in-place if possible
if numpy:
    or_i = numpy.ndarray.__ior__
else:
    def or_i(list0, list1):
        return [ x | y for x, y in zip(list0, list1) ]

def array_to_sprite_ega(self, byte_array, offset, dx, dy):
    """Build sprite from byte_array in EGA modes."""
    row_bytes = (dx+7) // 8
    attrs = []
    for y in range(dy):
        row = bytes_to_interval(byte_array[offset:offset+row_bytes], 8, 1)
        offset += row_bytes
        for plane in range(1, self.bitsperpixel):
            row = or_i(row, bytes_to_interval(byte_array[offset:offset+row_bytes], 8, 1 << plane))
            offset += row_bytes
        attrs.append(row[:dx])
    return attrs

def build_tile_cga(self, pattern):
    """Build a flood-fill tile for CGA screens."""
    tile = []
    bpp = self.bitsperpixel
    strlen = len(pattern)
    # in modes 1, (2), 3, 4, 5, 6 colours are encoded in consecutive bits
    # each byte represents one scan line
    mask = 8 - bpp
    for y in range(strlen):
        line = []
        for x in range(8): # width is 8//bpp
            c = 0
            for b in range(bpp-1, -1, -1):
                c = (c<<1) + ((pattern[y] >> (mask+b)) & 1)
            mask -= bpp
            if mask < 0:
                mask = 8 - bpp
            line.append(c)
        tile.append(line)
    return tile


class GraphicsMode(VideoMode):
    """Default settings for a graphics mode."""

    def __init__(
            self, name, pixel_width, pixel_height,
            text_height, text_width,
            attr, palette, colours, bitsperpixel,
            interleave_times, bank_size,
            num_pages=None,
            has_blink=False,
            supports_artifacts=False,
            cursor_index=None,
            pixel_aspect=None, aspect=None,
            video_segment=0xb800,
        ):
        """Initialise video mode settings."""
        font_width = int(pixel_width // text_width)
        font_height = int(pixel_height // text_height)
        self.interleave_times = int(interleave_times)
        # cga bank_size = 0x2000 interleave_times=2
        self.bank_size = int(bank_size)
        page_size = self.interleave_times * self.bank_size
        VideoMode.__init__(
            self, name, text_height, text_width,
            font_height, font_width, attr, palette, colours,
            num_pages, has_blink, video_segment, page_size
        )
        self.is_text_mode = False
        self.bitsperpixel = int(bitsperpixel)
        # number of pixels referenced in each byte of a plane
        self.ppb = 8 // self.bitsperpixel
        self.num_attr = 2**self.bitsperpixel
        self.bytes_per_row = int(pixel_width) * self.bitsperpixel // 8
        self.supports_artifacts = supports_artifacts
        self.cursor_index = cursor_index
        if pixel_aspect:
            self.pixel_aspect = pixel_aspect
        else:
            self.pixel_aspect = (self.pixel_height * aspect[0], self.pixel_width * aspect[1])

    def coord_ok(self, page, x, y):
        """Check if a page and coordinates are within limits."""
        return (
            page >= 0 and page < self.num_pages and
            x >= 0 and x < self.pixel_width and
            y >= 0 and y < self.pixel_height
        )

    def cutoff_coord(self, x, y):
        """Ensure coordinates are within screen + 1 pixel."""
        return min(self.pixel_width, max(-1, x)), min(self.pixel_height, max(-1, y))

    def set_plane(self, plane):
        """Set the current colour plane (EGA only)."""
        pass

    def set_plane_mask(self, mask):
        """Set the current colour plane mask (EGA only)."""
        pass


class CGAMode(GraphicsMode):
    """Default settings for a CGA graphics mode."""

    def get_coords(self, addr):
        """Get video page and coordinates for address."""
        addr = int(addr) - self.video_segment * 0x10
        # modes 1-5: interleaved scan lines, pixels sequentially packed into bytes
        page, addr = addr//self.page_size, addr%self.page_size
        # 2 x interleaved scan lines of 80bytes
        bank, offset = addr//self.bank_size, addr%self.bank_size
        row, col = offset//self.bytes_per_row, offset%self.bytes_per_row
        x = col * 8 // self.bitsperpixel
        y = bank + self.interleave_times * row
        return page, x, y

    def set_memory(self, screen, addr, byte_array):
        """Set bytes in CGA memory."""
        for page, x, y, ofs, length in walk_memory(self, addr, len(byte_array)):
            screen.drawing.put_interval(
                page, x, y, bytes_to_interval(byte_array[ofs:ofs+length], self.ppb)
            )

    def get_memory(self, screen, addr, num_bytes):
        """Retrieve bytes from CGA memory."""
        byte_array = bytearray(num_bytes)
        for page, x, y, ofs, length in walk_memory(self, addr, num_bytes):
            byte_array[ofs:ofs+length] = interval_to_bytes(
                screen.pixels.pages[page].get_interval(x, y, length*self.ppb), self.ppb
            )
        return byte_array

    def sprite_size_to_record(self, dx, dy):
        """Write 4-byte record of sprite size."""
        return struct.pack('<HH', dx*self.bitsperpixel, dy)

    def record_to_sprite_size(self, byte_array):
        """Read 4-byte record of sprite size."""
        w, dy = struct.unpack('<HH', byte_array[0:4])
        return w // self.bitsperpixel, dy

    def sprite_to_array(self, attrs, dx, dy, byte_array, offs):
        """Build the sprite byte array."""
        row_bytes = (dx * self.bitsperpixel + 7) // 8
        length = row_bytes*dy
        if offs+length > len(byte_array):
            # NOTE: if we use memoryviews instead of bytearrays, we won't need
            # this check as the assignment will fail with ValueError anyway
            raise ValueError('Sprite exceeds array byte size')
        byte_array[offs:offs+length] = b'\0'*length
        for row in attrs:
            byte_array[offs:offs+row_bytes] = interval_to_bytes(row, 8//self.bitsperpixel, 0)
            offs += row_bytes

    def array_to_sprite(self, byte_array, offset, dx, dy):
        """Build sprite from byte_array."""
        row_bytes = (dx * self.bitsperpixel + 7) // 8
        # illegal fn call if outside screen boundary
        attrs = []
        for y in range(dy):
            row = bytes_to_interval(byte_array[offset:offset+row_bytes], 8//self.bitsperpixel, 1)
            offset += row_bytes
            attrs.append(row[:dx])
        return attrs

    build_tile = build_tile_cga


class EGAMode(GraphicsMode):
    """Default settings for a EGA graphics mode."""

    def __init__(
            self, name, pixel_width, pixel_height,
            text_height, text_width,
            attr, palette, colours, bitsperpixel,
            interleave_times, bank_size, num_pages,
            colours1=None, has_blink=False, planes_used=range(4),
            aspect=None
        ):
        """Initialise video mode settings."""
        GraphicsMode.__init__(
            self, name, pixel_width, pixel_height,
            text_height, text_width,
            attr, palette, colours, bitsperpixel,
            interleave_times, bank_size,
            num_pages, has_blink, aspect=aspect
        )
        # EGA uses colour planes, 1 bpp for each plane
        self.ppb = 8
        self.bytes_per_row = pixel_width // 8
        self.video_segment = 0xa000
        self.planes_used = planes_used
        # additional colour plane mask
        self.master_plane_mask = sum([ 2**x for x in planes_used ])
        # this is a reference
        self.colours1 = colours1
        # current ega memory colour plane to read
        self.plane = 0
        # current ega memory colour planes to write to
        self.plane_mask = 0xff

    def set_plane(self, plane):
        """Set the current colour plane."""
        self.plane = plane

    def set_plane_mask(self, mask):
        """Set the current colour plane mask."""
        self.plane_mask = mask

    def get_coords(self, addr):
        """Get video page and coordinates for address."""
        addr = int(addr) - self.video_segment * 0x10
        # modes 7-9: 1 bit per pixel per colour plane
        page, addr = addr//self.page_size, addr%self.page_size
        x, y = (addr%self.bytes_per_row)*8, addr//self.bytes_per_row
        return page, x, y

    def get_memory(self, screen, addr, num_bytes):
        """Retrieve bytes from EGA memory."""
        plane = self.plane % (max(self.planes_used)+1)
        byte_array = bytearray(num_bytes)
        if plane not in self.planes_used:
            return byte_array
        for page, x, y, ofs, length in walk_memory(self, addr, num_bytes):
            byte_array[ofs:ofs+length] = interval_to_bytes(
                screen.pixels.pages[page].get_interval(x, y, length*self.ppb),
                self.ppb, plane
            )
        return byte_array

    def set_memory(self, screen, addr, byte_array):
        """Set bytes in EGA video memory."""
        # EGA memory is planar with memory-mapped colour planes.
        # Within a plane, 8 pixels are encoded into each byte.
        # The colour plane is set through a port OUT and
        # determines which bit of each pixel's attribute is affected.
        mask = self.plane_mask & self.master_plane_mask
        # return immediately for unused colour planes
        if mask == 0:
            return
        for page, x, y, ofs, length in walk_memory(self, addr, len(byte_array)):
            screen.drawing.put_interval(page, x, y,
                bytes_to_interval(byte_array[ofs:ofs+length], self.ppb, mask), mask
            )

    sprite_to_array = sprite_to_array_ega
    array_to_sprite = array_to_sprite_ega

    sprite_size_to_record = sprite_size_to_record_ega
    record_to_sprite_size = record_to_sprite_size_ega

    def build_tile(self, pattern):
        """Build a flood-fill tile."""
        tile = []
        bpp = self.bitsperpixel
        while len(pattern) % bpp != 0:
            # finish off the pattern with zeros
            pattern.append(0)
        strlen = len(pattern)
        # in modes (2), 7, 8, 9 each byte represents 8 bits
        # colour planes encoded in consecutive bytes
        mask = 7
        for y in range(strlen//bpp):
            line = []
            for x in range(8):
                c = 0
                for b in range(bpp-1, -1, -1):
                    c = (c<<1) + ((pattern[(y*bpp+b)%strlen] >> mask) & 1)
                mask -= 1
                if mask < 0:
                    mask = 7
                line.append(c)
            tile.append(line)
        return tile


class Tandy6Mode(GraphicsMode):
    """Default settings for Tandy graphics mode 6."""

    def __init__(self, *args, **kwargs):
        """Initialise video mode settings."""
        GraphicsMode.__init__(self, *args, **kwargs)
        # mode 6: 4x interleaved scan lines, 8 pixels per two bytes,
        # low attribute bits stored in even bytes, high bits in odd bytes.
        self.bytes_per_row = self.pixel_width * 2 // 8
        self.video_segment = 0xb800

    def get_coords(self, addr):
        """Get video page and coordinates for address."""
        addr =  int(addr) - self.video_segment * 0x10
        page, addr = addr//self.page_size, addr%self.page_size
        # 4 x interleaved scan lines of 160bytes
        bank, offset = addr//self.bank_size, addr%self.bank_size
        row, col = offset//self.bytes_per_row, offset%self.bytes_per_row
        x = (col // 2) * 8
        y = bank + 4 * row
        return page, x, y

    def get_memory(self, screen, addr, num_bytes):
        """Retrieve bytes from Tandy 640x200x4 """
        # 8 pixels per 2 bytes
        # low attribute bits stored in even bytes, high bits in odd bytes.
        half_len = (num_bytes+1) // 2
        hbytes = bytearray(half_len), bytearray(half_len)
        for parity in (0, 1):
            for page, x, y, ofs, length in walk_memory(self, addr, num_bytes, 2):
                hbytes[parity][ofs:ofs+length] = interval_to_bytes(
                    screen.pixels.pages[page].get_interval(x, y, length*self.ppb*2),
                    self.ppb*2, parity ^ (addr%2)
                )
        # resulting array may be too long by one byte, so cut to size
        return [item for pair in zip(*hbytes) for item in pair] [:num_bytes]

    def set_memory(self, screen, addr, byte_array):
        """Set bytes in Tandy 640x200x4 memory."""
        hbytes = byte_array[0::2], byte_array[1::2]
        # Tandy-6 encodes 8 pixels per byte, alternating colour planes.
        # I.e. even addresses are 'colour plane 0', odd ones are 'plane 1'
        for parity in (0, 1):
            mask = 2 ** (parity^(addr%2))
            for page, x, y, ofs, length in walk_memory(self, addr, len(byte_array), 2):
                screen.drawing.put_interval(
                    page, x, y,
                    bytes_to_interval(hbytes[parity][ofs:ofs+length], 2*self.ppb, mask),
                    mask
                )

    sprite_to_array = sprite_to_array_ega
    array_to_sprite = array_to_sprite_ega

    sprite_size_to_record = sprite_size_to_record_ega
    record_to_sprite_size = record_to_sprite_size_ega

    build_tile = build_tile_cga



"""
PC-BASIC - debug.py
DEBUG statement and utilities

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import io
import sys
import traceback
import logging
import platform
import struct
import tempfile
import subprocess
import importlib

from .base import error
from ..compat import PY2, WIN32, X64, BASE_DIR, which
from . import values
from . import api


def get_platform_info():
    """Show information about operating system and installed modules."""
    info = []
    info.append(u'\nPLATFORM')
    info.append(u'os: %s' % platform.platform())
    frozen = getattr(sys, 'frozen', '') or ''
    info.append(
        u'python: %s %s %s' % (
        sys.version.replace('\n', ''), ' '.join(platform.architecture()), frozen))
    info.append(u'\nMODULES')
    # try numpy before pygame to avoid strange ImportError on FreeBSD
    modules = ('numpy', 'pygame', 'curses', 'serial', 'parallel')
    for module in modules:
        try:
            m = importlib.import_module(module)
        except Exception:
            info.append(u'%s: --' % module)
        else:
            for version_attr in ('__version__', 'version', 'VERSION'):
                try:
                    name = module.split('.')[-1]
                    version = getattr(m, version_attr)
                    if isinstance(version, bytes):
                        version = version.decode('ascii', 'ignore')
                    info.append(u'%s: %s' % (name, version))
                    break
                except AttributeError:
                    pass
            else:
                info.append(u'%s: available' % module)
    info.append(u'\nLIBRARIES')
    try:
        from ..interface import video_sdl2
        info.append(u'sdl2: %s' % (video_sdl2.sdl2.get_dll_file()))
        if video_sdl2.sdlgfx:
            info.append(u'sdl2_gfx: %s' % (video_sdl2.sdlgfx.libfile))
        else:
            info.append(u'sdl2_gfx: --')
    except ImportError:
        info.append(u'sdl2: --')
        sdl2 = None
    info.append(u'\nEXTERNAL TOOLS')
    tools = (u'notepad', u'lpr', u'paps', u'beep', u'pbcopy', u'pbpaste')
    for tool in tools:
        location = which(tool) or u'--'
        info.append(u'%s: %s' % (tool, location))
    info.append(u'')
    return u'\n'.join(info)


class DebugException(BaseException):
    """Test exception for debugging purposes"""
    # inherit from BaseException to circumvent extension manager catching Exception

    def __repr__(self):
        return self.__doc__


class DebugSession(api.Session):
    """Debugging helper."""

    def __init__(self, *args, **kwargs):
        """Initialise debugger."""
        api.Session.__init__(self, *args, **kwargs)

    def start(self):
        """Start the session."""
        if not self._impl:
            api.Session.start(self)
            # register as an extension
            self._impl.extensions.add(self)
            # replace dummy debugging step
            self._impl.interpreter.step = self._debug_step
            self._do_trace = False
            self._watch_list = []

    def _debug_step(self, token):
        """Execute traces and watches on a program step."""
        outstr = u''
        if self._do_trace:
            linum = struct.unpack_from('<H', token, 2)
            outstr += u'[%i]' % linum
        for (expr, outs) in self._watch_list:
            outstr += u' %r = ' % (expr,)
            outs.seek(2)
            try:
                val = self._impl.parser.expression_parser.parse(outs)
                if isinstance(val, values.String):
                    outstr += u'"%s"' % self._impl.codepage.str_to_unicode(val.to_str())
                else:
                    outstr += values.to_repr(val, leading_space=False, type_sign=True)
            except Exception as e:
                logging.debug('%s %s', type(e), e)
                traceback.print_tb(sys.exc_info()[2])
        if outstr:
            logging.debug(outstr)

    def _handle_exception(self, e):
        """Handle exception during debugging."""
        logging.debug(b'%s %s', type(e), bytes(e))
        traceback.print_tb(sys.exc_info()[2])

    ###########################################################################
    # debugging commands

    def dir(self):
        """Show debugging commands."""
        logging.debug('Available commands:\n' + '\n'.join(
            '    _%s: %s' % (
                n.upper(), getattr(self, n).__doc__)
                for n in dir(self)
                    if '_' not in n and callable(getattr(self, n)) and n not in dir(api.Session)
            ))

    def crash(self):
        """Simulate a crash."""
        raise DebugException()

    def python(self, cmd):
        """Execute any Python code."""
        buf = io.BytesIO() if PY2 else io.StringIO()
        save_stdout = sys.stdout
        sys.stdout = buf
        try:
            exec(cmd)
        except Exception as e:
            self._handle_exception(e)
        sys.stdout = save_stdout
        logging.debug(buf.getvalue()[:-1]) # exclude \n

    def restart(self):
        """Ctrl+Alt+Delete."""
        raise error.Reset()

    def exit(self):
        """Quit the session."""
        raise error.Exit()

    def logprint(self, *args):
        """Write arguments to log."""
        logging.debug(self._impl.codepage.str_to_unicode(' '.join(bytes(arg) for arg in args)))

    def logwrite(self, *args):
        """Write arguments to log."""
        logging.debug(' '.join(repr(arg) for arg in args))

    def trace(self, on=True):
        """Switch line number tracing on or off."""
        self._do_trace = on

    def watch(self, expr):
        """Add an expression to the watch list."""
        outs = self._impl.tokeniser.tokenise_line(b'?' + expr)
        self._watch_list.append((expr, outs))

    def showvariables(self):
        """Dump all variables to the log."""
        repr_vars = '\n'.join((
            '==== Scalars ='.ljust(100, '='),
            repr(self._impl.scalars),
            '==== Arrays ='.ljust(100, '='),
            repr(self._impl.arrays),
            '==== Strings ='.ljust(100, '='),
            repr(self._impl.strings),
        ))
        for s in repr_vars.split('\n'):
            logging.debug(s)

    def showscreen(self):
        """Copy the screen buffer to the log."""
        for s in repr(self._impl.display.text_screen).split('\n'):
            logging.debug(self._impl.codepage.str_to_unicode(s))

    def showprogram(self):
        """Write a marked-up hex dump of the program to the log."""
        for s in repr(self._impl.program).split('\n'):
            logging.debug(s)

    def showplatform(self):
        """Show platform info."""
        logging.debug(get_platform_info())



"""
PC-BASIC - compat.posix
Interface for Unix-like system calls

(c) 2018--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

# pylint: disable=no-name-in-module

import os
import sys
import locale
import logging
import subprocess
import array

# set locale - this is necessary for curses and *maybe* for clipboard handling
# there's only one locale setting so best to do it all upfront here
# NOTE that this affects str.upper() etc.
# don't do this on Windows as it makes the console codepage different from the stdout encoding ?
try:
    locale.setlocale(locale.LC_ALL, '')
except locale.Error as e:
    # mis-configured locale can throw an error here, no need to crash
    logging.error(e)

from .base import PY2, HOME_DIR, MACOS

if PY2:
    from .python2 import which
else:
    from shutil import which


# text conventions
# ctrl+D
EOF = u'\x04'
# LF end-of-line
EOL = u'\n'

# shell conventions
# console encoding
SHELL_ENCODING = sys.stdin.encoding or locale.getpreferredencoding()
FS_ENCODING = sys.getfilesystemencoding()
# window suppression not needed on Unix
HIDE_WINDOW = None


# output buffer for ioctl call
_sock_size = array.array('i', [0])


##############################################################################
# various

def set_dpi_aware():
    """Enable HiDPI awareness."""

##############################################################################
# file system

def get_free_bytes(path):
    """Return the number of free bytes on the drive."""
    # note that from Python 3.3 onwards, we can use the cross-platform shutil.disk_usage
    st = os.statvfs(path)
    return st.f_bavail * st.f_frsize

def get_short_pathname(native_path):
    """Return Windows short path name or None if not available."""
    return None

if PY2:
    # the official parameter should be LC_CTYPE but that's None in my locale
    # on Windows, this would only work if the mbcs CP_ACP includes the characters we need;
    argv = [_arg.decode(FS_ENCODING, errors='replace') for _arg in sys.argv]

def is_hidden(path):
    """File is hidden."""
    # dot files are hidden on unixy systems
    base = os.path.basename(path)
    return base.startswith(u'.') and (base not in (u'.', u'..'))


##############################################################################
# printing

if which('paps'):
    def line_print(printbuf, printer):
        """Print the buffer to a LPR printer using PAPS."""
        options = b''
        if printer and printer != u'default':
            options = b'-P "%s"' % (printer.encode(SHELL_ENCODING, 'replace'),)
        if printbuf:
            # A4 paper is 595 points wide by 842 points high.
            # Letter paper is 612 by 792 points.
            # the below seems to allow 82 chars horizontally on A4; it appears
            # my PAPS version doesn't quite use cpi correctly as 10cpi should
            # allow 80 chars on A4 with a narrow margin but only does so with a
            # margin of 0.
            pr = subprocess.Popen(
                b'paps --cpi=11 --lpi=6 --left-margin=20 --right-margin=20 '
                b'--top-margin=6 --bottom-margin=6 '
                b'| lpr %s' % (options,), shell=True, stdin=subprocess.PIPE)
            # PAPS does not recognise CRLF
            printbuf = printbuf.replace(b'\r\n', b'\n')
            pr.stdin.write(printbuf)
            pr.stdin.close()

else:
    def line_print(printbuf, printer):
        """Print the buffer to a LPR (CUPS or older UNIX) printer."""
        options = b''
        if printer and printer != u'default':
            options = b'-P "%s"' % (printer.encode(SHELL_ENCODING, 'replace'),)
        if printbuf:
            # cups defaults to 10 cpi, 6 lpi.
            pr = subprocess.Popen(b'lpr %s' % (options,), shell=True, stdin=subprocess.PIPE)
            pr.stdin.write(printbuf)
            pr.stdin.close()



"""
PC-BASIC - audio_sdl2.py
Sound interface based on SDL2

(c) 2015--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

# see e.g. http://toomanyideas.net/2014/pysdl2-playing-a-sound-from-a-wav-file.html

import logging
import ctypes
from collections import deque

try:
    from . import sdl2
except ImportError:
    sdl2 = None

try:
    import numpy
except ImportError:
    numpy = None

from .audio import AudioPlugin
from .base import audio_plugins, InitFailed
from . import synthesiser


# approximate generator chunk length
# one wavelength at 37 Hz is 1192 samples at 44100 Hz
CHUNK_LENGTH = 1192 * 4
# length of chunks to be consumed by callback
CALLBACK_CHUNK_LENGTH = 2048
# number of samples below which to replenish the buffer
MIN_SAMPLES_BUFFER = 2*CALLBACK_CHUNK_LENGTH


##############################################################################
# plugin

@audio_plugins.register('sdl2')
class AudioSDL2(AudioPlugin):
    """SDL2-based audio plugin."""

    def __init__(self, audio_queue, **kwargs):
        """Initialise sound system."""
        if not sdl2:
            raise InitFailed('Module `sdl2` not found')
        if not numpy:
            raise InitFailed('Module `numpy` module not found')
        # synthesisers
        self.signal_sources = synthesiser.get_signal_sources()
        # sound generators for each voice
        self.generators = [deque(), deque(), deque(), deque()]
        # buffer of samples; drained by callback, replenished by _play_sound
        self.samples = [numpy.array([], numpy.int16) for _ in range(4)]
        # SDL AudioDevice and specifications
        self.audiospec = sdl2.SDL_AudioSpec(0, 0, 0, 0)
        self.audiospec.freq = synthesiser.SAMPLE_RATE
        # samples are 16-bit signed ints
        self.audiospec.format = sdl2.AUDIO_S16SYS
        self.audiospec.channels = 1
        self.audiospec.samples = CALLBACK_CHUNK_LENGTH
        self.audiospec.callback = sdl2.SDL_AudioCallback(self._get_next_chunk)
        self.dev = None
        AudioPlugin.__init__(self, audio_queue)

    def __enter__(self):
        """Perform any necessary initialisations."""
        # init sdl audio in this thread separately
        sdl2.SDL_Init(sdl2.SDL_INIT_AUDIO)
        self.dev = sdl2.SDL_OpenAudioDevice(None, 0, self.audiospec, None, 0)
        if self.dev == 0:
            logging.warning('Could not open audio device: %s', sdl2.SDL_GetError())
        # unpause the audio device
        sdl2.SDL_PauseAudioDevice(self.dev, 0)
        return AudioPlugin.__enter__(self)

    def tone(self, voice, frequency, duration, loop, volume):
        """Enqueue a tone."""
        self.generators[voice].append(synthesiser.SoundGenerator(
            self.signal_sources[voice], synthesiser.FEEDBACK_TONE,
            frequency, duration, loop, volume
        ))

    def noise(self, source, frequency, duration, loop, volume):
        """Enqueue a noise."""
        feedback = synthesiser.FEEDBACK_NOISE if source else synthesiser.FEEDBACK_PERIODIC
        self.generators[3].append(synthesiser.SoundGenerator(
            self.signal_sources[3], feedback,
            frequency, duration, loop, volume
        ))

    def hush(self):
        """Stop sound."""
        for voice in range(4):
            self._next_tone[voice] = None
            while self.generators[voice]:
                self.generators[voice].popleft()
        sdl2.SDL_LockAudioDevice(self.dev)
        self.samples = [numpy.array([], numpy.int16) for _ in range(4)]
        sdl2.SDL_UnlockAudioDevice(self.dev)

    def _work(self):
        """Replenish sample buffer."""
        for voice in range(4):
            if len(self.samples[voice]) > MIN_SAMPLES_BUFFER:
                # nothing to do
                continue
            while True:
                if self._next_tone[voice] is None or self._next_tone[voice].loop:
                    try:
                        # looping tone will be interrupted
                        # by any new tone appearing in the generator queue
                        self._next_tone[voice] = self.generators[voice].popleft()
                    except IndexError:
                        if self._next_tone[voice] is None:
                            current_chunk = None
                            break
                current_chunk = self._next_tone[voice].build_chunk(CHUNK_LENGTH)
                if current_chunk is not None:
                    break
                self._next_tone[voice] = None
            if current_chunk is not None:
                # append chunk to samples list
                # lock to ensure callback doesn't try to access the list too
                sdl2.SDL_LockAudioDevice(self.dev)
                self.samples[voice] = numpy.concatenate((self.samples[voice], current_chunk))
                sdl2.SDL_UnlockAudioDevice(self.dev)

    def _get_next_chunk(self, notused, stream, length_bytes):
        """Callback function to generate the next chunk to be played."""
        # this is for 16-bit samples
        length = length_bytes // 2
        samples = [self.samples[voice][:length] for voice in range(4)]
        self.samples = [self.samples[voice][length:] for voice in range(4)]
        # if samples have run out, add silence
        for voice in range(4):
            if len(samples[voice]) < length:
                silence = numpy.zeros(length-len(samples[voice]), numpy.int16)
                samples[voice] = numpy.concatenate((samples[voice], silence))
        # mix the samples by averaging
        # we need the int32 intermediate step, for int16 numpy will average [32767, 32767] to -1
        mixed = bytearray(
            numpy.array(numpy.mean(samples, axis=0, dtype=numpy.int32), dtype=numpy.int16).data
        )
        ctypes.memmove(
            stream, (ctypes.c_char * length_bytes).from_buffer(mixed[:length_bytes]), length_bytes
        )



"""
PC-BASIC - api.py
Session API

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os

from ..compat import text_type

from .base import error
from .devices import NameWrapper
from . import implementation

from ..data import read_codepage as codepage
from ..data import read_fonts as font


class Session(object):
    """Public API to BASIC session."""

    def __init__(self, interface=None, **kwargs):
        """Set up session object."""
        self.interface = interface
        self._kwargs = kwargs
        self._impl = None

    def __enter__(self):
        """Context guard."""
        return self

    def __exit__(self, ex_type, ex_val, tb):
        """Context guard."""
        self.close()
        # catch Exit and Break events
        if ex_type in (error.Exit, error.Break):
            return True

    def __getstate__(self):
        """Pickle the session."""
        pickle_dict = self.__dict__.copy()
        pickle_dict['interface'] = None
        return pickle_dict

    def __setstate__(self, pickle_dict):
        """Unpickle and resume the session."""
        self.__dict__.update(pickle_dict)

    def start(self):
        """Start the session."""
        if not self._impl:
            self._impl = implementation.Implementation(**self._kwargs)
            self._impl.attach_interface(self.interface)

    def attach(self, interface=None):
        """Attach interface to interpreter session."""
        self.start()
        self.interface = interface
        self._impl.attach_interface(interface)
        return self

    def bind_file(self, file_name_or_object, name=None, create=False):
        """Bind a native file name or Python stream to a BASIC file name."""
        self.start()
        # if a file name, resolve
        if (
                not isinstance(file_name_or_object, (bytes, text_type))
                or os.path.isfile(file_name_or_object)
            ):
            # if it's an object or the file name exists, use it
            return self._impl.files.get_device(b'@:').bind(file_name_or_object, name)
        elif create and (
                not os.path.dirname(file_name_or_object) or
                os.path.isdir(os.path.dirname(file_name_or_object))
            ):
            # if it doesn't and we're allowed to create and the directory exists, create new
            return self._impl.files.get_device(b'@:').bind(file_name_or_object, name)
        # not resolved, try to use/create as internal name
        return NameWrapper(self._impl.codepage, file_name_or_object)

    def execute(self, command):
        """Execute a BASIC statement."""
        self.start()
        with self._impl.io_streams.activate():
            for cmd in command.splitlines():
                if isinstance(cmd, text_type):
                    cmd = self._impl.codepage.str_from_unicode(cmd)
                self._impl.execute(cmd)

    def evaluate(self, expression):
        """Evaluate a BASIC expression."""
        self.start()
        with self._impl.io_streams.activate():
            if isinstance(expression, text_type):
                expression = self._impl.codepage.str_from_unicode(expression)
            return self._impl.evaluate(expression)

    def set_variable(self, name, value):
        """Set a variable in memory."""
        self.start()
        if isinstance(name, text_type):
            name = name.encode('ascii')
        name = name.upper()
        self._impl.set_variable(name, value)

    def get_variable(self, name, as_type=None):
        """Get a variable in memory."""
        self.start()
        if isinstance(name, text_type):
            name = name.encode('ascii')
        return self._impl.get_variable(name, as_type)

    def convert(self, value, to_type):
        """Convert a Python value to another type, consistent with BASIC rules."""
        self.start()
        return self._impl.get_converter(type(value), to_type)(value)

    def interact(self):
        """Interactive interpreter session."""
        self.start()
        with self._impl.io_streams.activate():
            self._impl.interact()

    def close(self):
        """Close the session."""
        if self._impl:
            self._impl.close()



"""
PC-BASIC - guard
Crash guard

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import sys
import logging
import platform
import tempfile
import traceback
import json
from datetime import datetime
from contextlib import contextmanager
from subprocess import check_output, CalledProcessError

from .metadata import VERSION
from .basic.base import error, signals
from .data import get_data, ResourceFailed


LOG_PATTERN = u'crash-%Y%m%d-'
PAUSE_MESSAGE = u'Fatal error. Press a key to close this window'


class NoGuard(object):
    """Null context manager."""

    @contextmanager
    def protect(self, *args, **kwargs):
        yield

NOGUARD = NoGuard()

try:
    RELEASE_ID = json.loads(get_data('release.json'))
    TAG = RELEASE_ID[u'tag']
    COMMIT = RELEASE_ID[u'commit']
    TIMESTAMP = RELEASE_ID[u'timestamp']
except ResourceFailed:
    TAG = u''
    TIMESTAMP = u''
    try:
        COMMIT = check_output(
            ['git', 'describe', '--always'], cwd=os.path.dirname(__file__)
        ).strip().decode('ascii', 'ignore')
    except (CalledProcessError, EnvironmentError):
        COMMIT = u'unknown'


class ExceptionGuard(object):
    """Context manager to handle uncaught exceptions."""

    def __init__(self, log_dir=u'', uargv=()):
        """Initialise crash guard."""
        self._uargv = uargv
        self._log_dir = log_dir

    @contextmanager
    def protect(self, interface, session):
        """Crash context guard."""
        try:
            yield
        except (error.Exit, error.Reset):
            raise
        except BaseException:
            if not self._bluescreen(session._impl, interface, *sys.exc_info()):
                raise
            interface.pause(PAUSE_MESSAGE)

    def _bluescreen(self, impl, iface, exc_type, exc_value, exc_traceback):
        """Display modal message"""
        if not impl:
            return False
        if iface:
            iface_name = u'%s, %s' % (type(iface._video).__name__, type(iface._audio).__name__)
        else:
            iface_name = u'--'
        # log the standard python error
        stack = traceback.extract_tb(exc_traceback)
        logging.error(''.join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
        # obtain statement being executed
        if impl.interpreter.run_mode:
            codestream = impl.program.bytecode
            bytepos = codestream.tell() - 1
            from_line = impl.program.get_line_number(bytepos)
            try:
                codestream.seek(impl.program.line_numbers[from_line]+1)
                _, output, textpos = impl.lister.detokenise_line(codestream, bytepos)
                code_line = bytes(output)
            except KeyError:
                code_line = b'<could not retrieve line number %d>' % from_line
        else:
            impl.interpreter.direct_line.seek(0)
            code_line = bytes(
                impl.lister.detokenise_compound_statement(impl.interpreter.direct_line)[0]
            )
        # don't risk codepage logic here, use cp437
        code_line = code_line.decode('cp437', 'replace')
        # stop program execution
        impl.interpreter.set_pointer(False)
        # create crash log file
        logname = datetime.now().strftime(LOG_PATTERN)
        logfile = tempfile.NamedTemporaryFile(
            mode='wb', suffix='.log', prefix=logname, dir=self._log_dir, delete=False
        )
        # construct the message
        frozen = getattr(sys, 'frozen', u'') or u''
        message = [
            (0x70, u'FATAL ERROR\n'),
            (0x17, u'version   '),
            (0x1f, u'%s [%s] %s %s' % (VERSION, COMMIT, TAG, TIMESTAMP)),
            (0x17, u'\npython    '),
            (0x1f, u'%s [%s] %s' % (
                platform.python_version(), u' '.join(platform.architecture()), frozen
            )),
            (0x17, u'\nplatform  '),
            (0x1f, platform.platform()),
            (0x17, u'\ninterface '),
            (0x1f, iface_name),
            (0x17, u'\nstatement '),
            (0x1f, code_line + u'\n\n'),
        ] + [
            (0x1f, u'{0}:{1}, {2}\n'.format(os.path.split(s[0])[-1], s[1], s[2]))
            for s in stack[-4:]
        ] + [
            (0x1f, u'{0}:'.format(exc_type.__name__)),
            (0x17, u' {0}\n\n'.format(exc_value)),
            (0x70, u'This is a bug in PC-BASIC.\n'),
            (0x17, u'Sorry about that. '),
            (0x17, u'To help improve PC-BASIC, '),
            (0x70, u'please file a bug report'),
            (0x17, u' at\n  '),
            (0x1f, u'https://github.com/robhagemans/pcbasic/issues'),
            (0x17, u'\nPlease include the messages above and '),
            (0x17, u'as much information as you can about what you were doing and '),
            (0x17, u'how this happened. Please attach the log file\n  '),
            (0x1f, logfile.name),
            (0x17, u'\nThank you!\n\n'),
            (0x17, u'Press a key to close this window.'),
        ]
        bottom = (0x70,
            u'This message has been copied onto the clipboard. You can paste it with Ctrl-V.'
        )
        # create crash log
        crashlog = [
            u'PC-BASIC crash log',
            u'=' * 100,
            u''.join(text for _, text in message),
            u''.join(traceback.format_exception(exc_type, exc_value, exc_traceback)),
            u'==== Screen Pages ='.ljust(100, u'='),
            repr(impl.display.text_screen),
            u'==== Scalars ='.ljust(100, u'='),
            repr(impl.scalars),
            u'==== Arrays ='.ljust(100, u'='),
            repr(impl.arrays),
            u'==== Strings ='.ljust(100, u'='),
            repr(impl.strings),
            u'==== Program Buffer ='.ljust(100, u'='),
            repr(impl.program),
        ]
        impl.program.bytecode.seek(1)
        crashlog.append(u'==== Program ='.ljust(100, u'='))
        while True:
            _, line, _ = impl.lister.detokenise_line(impl.program.bytecode)
            if not line:
                break
            crashlog.append(bytes(line).decode('cp437', 'replace'))
        crashlog.append(u'==== Options ='.ljust(100, u'='))
        crashlog.append(repr(self._uargv))
        # clear screen for modal message
        # choose attributes - this should be readable on VGA, MDA, PCjr etc.
        impl.display.screen(0, 0, 0, 0, new_width=80)
        impl.display.set_attr(0x17)
        impl.display.set_border(1)
        impl.display.text_screen.clear()
        impl.display.text_screen._bottom_row_allowed = True
        # show message on screen
        for attr, text in message:
            impl.display.set_attr(attr)
            impl.display.text_screen.write(text.replace('\n', '\r').encode('cp437', 'replace'))
        impl.display.text_screen.set_pos(25, 1)
        impl.display.set_attr(bottom[0])
        impl.display.text_screen.write(bottom[1].replace('\n', '\r').encode('cp437', 'replace'))
        # write crash log
        crashlog = u'\n'.join(
            line.decode('cp437', 'replace') if isinstance(line, bytes) else line
            for line in crashlog
        )
        with logfile as f:
            f.write(crashlog.encode('utf-8', 'replace'))
        # put on clipboard
        impl.queues.video.put(signals.Event(signals.VIDEO_SET_CLIPBOARD_TEXT, (crashlog, False)))
        return True



"""
PC-BASIC - compat
Cross-platform compatibility utilities

(c) 2018--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

# pylint: disable=no-name-in-module, import-error, undefined-variable, used-before-assignment

import sys
import os

from .base import PLATFORM, PY2, WIN32, MACOS, X64
from .base import USER_CONFIG_HOME, USER_DATA_HOME, BASE_DIR, HOME_DIR
from .base import split_quoted, muffle


if PY2:
    from .python2 import add_str, iterchar
    from .python2 import xrange, zip, iteritems, itervalues, iterkeys
    from .python2 import getcwdu, getenvu, setenvu, iterenvu
    from .python2 import configparser, queue, copyreg, which
    from .python2 import SimpleNamespace
    unichr, int2byte, text_type = unichr, chr, unicode

    if WIN32:
        from . import win32_subprocess
        from .win32 import argv
    else:
        from .posix import argv
else:
    import configparser, queue, copyreg
    from shutil import which
    from types import SimpleNamespace
    from .python3 import int2byte, add_str, iterchar
    from .python3 import xrange, zip, iteritems, itervalues, iterkeys
    from .python3 import getcwdu, getenvu, setenvu, iterenvu
    unichr, text_type = chr, str
    argv = sys.argv


if WIN32:
    from .win32_console import console, read_all_available, stdin, stdout, stderr, IS_CONSOLE_APP
    from .win32 import set_dpi_aware, line_print
    from .win32 import get_free_bytes, get_short_pathname, is_hidden
    from .win32 import EOL, EOF
    from .win32 import SHELL_ENCODING, HIDE_WINDOW
else:
    from .posix_console import console, read_all_available, stdin, stdout, stderr, IS_CONSOLE_APP
    from .posix import set_dpi_aware, line_print
    from .posix import get_free_bytes, get_short_pathname, is_hidden
    from .posix import EOL, EOF
    from .posix import SHELL_ENCODING, HIDE_WINDOW


if MACOS:
    # on MacOS, if launched from Finder, ignore the additional "process serial number" argument
    argv = [_arg for _arg in argv if not _arg.startswith(b'-psn_')]
    # for macOS - if no console, presumably we're launched as a bundle
    # set working directory to user home
    # bit of a hack but I don't know a better way
    if not IS_CONSOLE_APP:
        os.chdir(HOME_DIR)



"""
PC-BASIC - font.py
Font handling

(c) 2014--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import logging
import binascii

try:
    import numpy
except ImportError:
    numpy = None

from ...compat import iteritems, int2byte

from ..base import signals


# ascii codepoints for which to repeat column 8 in column 9 (box drawing)
# Many internet sources say this should be 0xC0--0xDF. However, that would
# exclude the shading characters. It appears to be traced back to a mistake in
# IBM's VGA docs. See https://01.org/linuxgraphics/sites/default/files/documentation/ilk_ihd_os_vol3_part1r2.pdf
CARRY_COL_9_CHARS = tuple(int2byte(_c) for _c in range(0xb0, 0xdf+1))
# ascii codepoints for which to repeat row 8 in row 9 (box drawing)
CARRY_ROW_9_CHARS = tuple(int2byte(_c) for _c in range(0xb0, 0xdf+1))


# The glyphs below are extracted from Henrique Peron's CPIDOS v3.0,
# CPIDOS is distributed with FreeDOS at
#   http://www.freedos.org/software/?prog=cpidos
#   http://www.ibiblio.org/pub/micro/pc-stuff/freedos/files/dos/cpi/
# CPIDOS is Copyright (C) 2002-2011 by Henrique Peron (hperon@terra.com.br)
# and licensed under the GNU GPL version 2 or later.
DEFAULT_FONT = {
    int2byte(_i): binascii.unhexlify(_v) for _i, _v in enumerate((
    '0000000000000000', '7e81a581bd99817e', '7effdbffc3e7ff7e', '6cfefefe7c381000',
    '10387cfe7c381000', '387c38fefed61038', '10387cfefe7c1038', '0000183c3c180000',
    'ffffe7c3c3e7ffff', '003c664242663c00', 'ffc399bdbd99c3ff', '0f070f7dcccccc78',
    '3c6666663c187e18', '3f333f303070f0e0', '7f637f636367e6c0', '18db3ce7e73cdb18',
    '80e0f8fef8e08000', '020e3efe3e0e0200', '183c7e18187e3c18', '6666666666006600',
    '7fdbdb7b1b1b1b00', '3e613c66663c867c', '000000007e7e7e00', '183c7e187e3c18ff',
    '183c7e1818181800', '181818187e3c1800', '00180cfe0c180000', '003060fe60300000',
    '0000c0c0c0fe0000', '002466ff66240000', '00183c7effff0000', '00ffff7e3c180000',
    '0000000000000000', '00183c3c18180018', '0066662400000000', '006c6cfe6cfe6c6c',
    '00183e603c067c18', '0000c6cc183066c6', '00386c3876dccc76', '0018183000000000',
    '000c18303030180c', '0030180c0c0c1830', '0000663cff3c6600', '000018187e181800',
    '0000000000181830', '000000007e000000', '0000000000001818', '00060c183060c080',
    '00386cc6d6c66c38', '001838181818187e', '007cc6061c3066fe', '007cc6063c06c67c',
    '001c3c6cccfe0c1e', '00fec0c0fc06c67c', '003860c0fcc6c67c', '00fec60c18303030',
    '007cc6c67cc6c67c', '007cc6c67e060c78', '0000181800001818', '0018180000181830',
    '00060c1830180c06', '0000007e00007e00', '006030180c183060', '007cc60c18180018',
    '007cc6dededec078', '00386cc6fec6c6c6', '00fc66667c6666fc', '003c66c0c0c0663c',
    '00f86c6666666cf8', '00fe6268786862fe', '00fe6268786860f0', '003c66c0c0ce663a',
    '00c6c6c6fec6c6c6', '003c18181818183c', '001e0c0c0ccccc78', '00e6666c786c66e6',
    '00f06060606266fe', '00c6eefefed6c6c6', '00c6e6f6decec6c6', '007cc6c6c6c6c67c',
    '00fc66667c6060f0', '007cc6c6c6ce7c0e', '00fc66667c6c66e6', '003c6630180c663c',
    '007e7e5a1818183c', '00c6c6c6c6c6c67c', '00c6c6c6c6c66c38', '00c6c6c6d6d6fe6c',
    '00c6c66c386cc6c6', '006666663c18183c', '00fec68c183266fe', '003c30303030303c',
    '00c06030180c0602', '003c0c0c0c0c0c3c', '10386c0000000000', '00000000000000ff',
    '30180c0000000000', '000000780c7ccc76', '00e0607c666666dc', '0000007cc6c0c67c',
    '001c0c7ccccccc76', '0000007cc6fec07c', '003c6660f86060f0', '00000076cc7c0cf8',
    '00e0606c766666e6', '001800381818183c', '000600060606663c', '00e060666c786ce6',
    '003818181818183c', '000000ecfed6d6d6', '000000dc66666666', '0000007cc6c6c67c',
    '000000dc667c60f0', '00000076cc7c0c1e', '000000dc766060f0', '0000007ec07c06fc',
    '003030fc3030361c', '000000cccccccc76', '000000c6c6c66c38', '000000c6d6d6fe6c',
    '000000c66c386cc6', '000000c6c67e06fc', '0000007e0c18307e', '000e18187018180e',
    '0018181818181818', '007018180e181870', '76dc000000000000', '000010386cc6c6fe',
    '3c66c0c0663c1870', '00cc00cccccccc76', '0c18007cc6fec07c', '386c00780c7ccc76',
    '00cc00780c7ccc76', '603000780c7ccc76', '386c38780c7ccc76', '007cc6c0c67c1870',
    '386c007cc6fec07c', '00c6007cc6fec07c', '3018007cc6fec07c', '006600381818183c',
    '386c00381818183c', '301800381818183c', 'c610386cc6fec6c6', '386c387cc6fec6c6',
    '0c18fec0f8c0c0fe', '000000ec367ed86e', '003e6cccfeccccce', '386c007cc6c6c67c',
    '00c6007cc6c6c67c', '3018007cc6c6c67c', '78cc00cccccccc76', '603000cccccccc76',
    '00c600c6c67e06fc', 'c600386cc6c66c38', 'c600c6c6c6c6c67c', '18187ec0c07e1818',
    '386c64f0606066fc', '66663c7e187e1818', 'f8ccccfac6cfc6c7', '0e1b183c1818d870',
    '183000780c7ccc76', '0c1800381818183c', '0c18007cc6c6c67c', '183000cccccccc76',
    '76dc00dc66666666', '76dc00e6f6decec6', '003c6c6c36007e00', '00386c6c38007c00',
    '003000303060c67c', '000000fec0c00000', '000000fe06060000', '63e66c7e3366cc0f',
    '63e66c7a366adf06', '00180018183c3c18', '0000003366cc6633', '000000cc663366cc',
    '2288228822882288', '55aa55aa55aa55aa', '77dd77dd77dd77dd', '1818181818181818',
    '18181818f8181818', '1818f818f8181818', '36363636f6363636', '00000000fe363636',
    '0000f818f8181818', '3636f606f6363636', '3636363636363636', '0000fe06f6363636',
    '3636f606fe000000', '36363636fe000000', '1818f818f8000000', '00000000f8181818',
    '181818181f000000', '18181818ff000000', '00000000ff181818', '181818181f181818',
    '00000000ff000000', '18181818ff181818', '18181f181f181818', '3636363637363636',
    '363637303f000000', '00003f3037363636', '3636f700ff000000', '0000ff00f7363636',
    '3636373037363636', '0000ff00ff000000', '3636f700f7363636', '1818ff00ff000000',
    '36363636ff000000', '0000ff00ff181818', '00000000ff363636', '363636363f000000',
    '18181f181f000000', '00001f181f181818', '000000003f363636', '36363636ff363636',
    '1818ff18ff181818', '18181818f8000000', '000000001f181818', 'ffffffffffffffff',
    '00000000ffffffff', 'f0f0f0f0f0f0f0f0', '0f0f0f0f0f0f0f0f', 'ffffffff00000000',
    '00000076dcc8dc76', '78ccd8ccc6c6cc00', '00fe6260606060f0', '000000fe6c6c6c6c',
    '00fec6603060c6fe', '0000007ed8d8d870', '0000006666667cc0', '000000fe3030361c',
    '00107cd6d6d67c10', '007cc6c6fec6c67c', '00386cc6c66c28ee', '003c60387cc6c67c',
    '00007edbdb7e0000', '0000005cd6d67c10', '0000007cc670c67c', '007cc6c6c6c6c600',
    '0000fe00fe00fe00', '0018187e1818007e', '0030180c1830007e', '000c1830180c007e',
    '0e1b1b1818181818', '1818181818d8d870', '000018007e001800', '000076dc0076dc00',
    '00386c6c38000000', '0000000018180000', '0000000018000000', '0f0c0c0cec6c3c1c',
    '006c363636360000', '00780c18307c0000', '00003c3c3c3c0000', '0000000000000000',
))}


class Font(object):
    """Single-height bitfont."""

    def __init__(self, height=8, fontdict=None):
        """Initialise the font."""
        self._height = height
        if height == 8 and not fontdict:
            fontdict = DEFAULT_FONT
        self._fontdict = fontdict

    def get_byte(self, charvalue, offset):
        """Get byte sequency for character."""
        return ord(self._fontdict[int2byte(charvalue)][offset])

    def set_byte(self, charvalue, offset, byte):
        """Set byte sequency for character."""
        old = self._fontdict[int2byte(charvalue)]
        self._fontdict[int2byte(charvalue)] = old[:offset%8] + byte + old[offset%8+1:]
        #self.screen.rebuild_glyph(charvalue)

    def build_glyph(self, c, req_width, req_height):
        """Build a glyph for the given codepage character."""
        try:
            face = bytearray(self._fontdict[c])
        except KeyError:
            logging.debug('Code point [%r] not represented in font, replacing with blank glyph.', c)
            face = bytearray(int(self._height))
        # shape of encoded mask (8 or 16 wide; usually 8, 14 or 16 tall)
        code_height = 8 if req_height == 9 else req_height
        code_width = (8 * len(face)) // code_height
        force_double = req_width >= code_width * 2
        force_single = code_width >= (req_width-1) * 2
        if force_double or force_single:
            # i.e. we need a double-width char but got single or v.v.
            logging.debug(
                'Incorrect glyph width for code point [%r]: %d-pixel requested, %d-pixel found.',
                c, req_width, code_width
            )
        return _unpack_glyph(
            face, code_height, code_width, req_height, req_width,
            force_double, force_single,
            c in CARRY_COL_9_CHARS, c in CARRY_ROW_9_CHARS
        )

if numpy:

    def _unpack_glyph(
            face, code_height, code_width, req_height, req_width,
            force_double, force_single, carry_col_9, carry_row_9
        ):
        """Convert byte list to glyph pixels, numpy implementation."""
        glyph = numpy.unpackbits(face, axis=0).reshape((code_height, code_width)).astype(bool)
        # repeat last rows (e.g. for 9-bit high chars)
        if req_height > glyph.shape[0]:
            if carry_row_9:
                repeat_row = glyph[-1]
            else:
                repeat_row = numpy.zeros((1, code_width), dtype=numpy.uint8)
            while req_height > glyph.shape[0]:
                glyph = numpy.vstack((glyph, repeat_row))
        if force_double:
            glyph = glyph.repeat(2, axis=1)
        elif force_single:
            glyph = glyph[:, ::2]
        # repeat last cols (e.g. for 9-bit wide chars)
        if req_width > glyph.shape[1]:
            if carry_col_9:
                repeat_col = numpy.atleast_2d(glyph[:,-1]).T
            else:
                repeat_col = numpy.zeros((code_height, 1), dtype=numpy.uint8)
            while req_width > glyph.shape[1]:
                glyph = numpy.hstack((glyph, repeat_col))
        return glyph

else:

    def _unpack_glyph(
            face, code_height, code_width, req_height, req_width,
            force_double, force_single, carry_col_9, carry_row_9
        ):
        """Convert byte list to glyph pixels, non-numpy implementation."""
        # req_width can be 8, 9 (SBCS), 16, 18 (DBCS) only
        req_width_base = req_width if req_width <= 9 else req_width // 2
        # if our code glyph is too wide for request, we need to make space
        start_width = req_width*2 if force_single else req_width
        glyph = [ [False]*start_width for _ in range(req_height) ]
        for yy in range(code_height):
            for half in range(code_width//8):
                line = face[yy*(code_width//8)+half]
                for xx in range(8):
                    if (line >> (7-xx)) & 1 == 1:
                        glyph[yy][half*8 + xx] = True
            # halve the width if code width incorrect
            if force_single:
                glyph[yy] = glyph[yy][::2]
            # MDA/VGA 9-bit characters
            # carry_col_9 will be ignored for double-width glyphs
            if carry_col_9 and req_width == 9:
                glyph[yy][8] = glyph[yy][7]
        # tandy 9-bit high characters
        if carry_row_9 and req_height == 9:
            for xx in range(8):
                glyph[8][xx] = glyph[7][xx]
        # double the width if code width incorrect
        if force_double:
            for yy in range(code_height):
                for xx in range(req_width_base-1, -1, -1):
                    glyph[yy][2*xx+1] = glyph[yy][xx]
                    glyph[yy][2*xx] = glyph[yy][xx]
        return glyph


#######################################################################################
# glyph cache

class GlyphCache(object):

    def __init__(self, mode, fonts, codepage, queues):
        """Initialise glyph set."""
        self._queues = queues
        self._mode = mode
        self._fonts = fonts
        self._codepage = codepage
        # preload SBCS glyphs
        self._glyphs = {
            c: self._fonts[mode.font_height].build_glyph(c, mode.font_width, mode.font_height)
            for c in map(int2byte, range(256))
        }
        self.submit()

    def submit(self):
        """Send glyph dict to interface."""
        if self._mode.is_text_mode:
            # send glyphs to signals; copy is necessary
            # as dict may change here while the other thread is working on it
            self._queues.video.put(signals.Event(
                signals.VIDEO_BUILD_GLYPHS, (
                    {self._codepage.to_unicode(k, u'\0'): v for k, v in iteritems(self._glyphs)},
                )
            ))

    def rebuild_glyph(self, ordval):
        """Rebuild a text-mode character after POKE."""
        if self._mode.is_text_mode:
            # force rebuilding the character by deleting and requesting
            del self._glyphs[int2byte(ordval)]
            self._submit_char(int2byte(ordval))

    def _submit_char(self, char):
        """Rebuild glyph and send to interface."""
        mask = self._fonts[self._mode.font_height].build_glyph(
            char, self._mode.font_width*2, self._mode.font_height
        )
        self._glyphs[char] = mask
        if self._mode.is_text_mode:
            self._queues.video.put(signals.Event(
                signals.VIDEO_BUILD_GLYPHS, ({self._codepage.to_unicode(char, u'\0'): mask},)
            ))

    def check_char(self, char):
        """Submit glyph if needed."""
        if self._mode.is_text_mode and char not in self._glyphs:
            self._submit_char(char)

    if numpy:
        def get_sprite(self, row, col, char, fore, back):
            """Return a sprite for a given character."""
            if char not in self._glyphs:
                self._submit_char(char)
            mask = self._glyphs[char]
            # set background
            glyph = numpy.full(mask.shape, back, dtype=int)
            # stamp foreground mask
            glyph[mask] = fore
            x0, y0 = (col-1) * self._mode.font_width, (row-1) * self._mode.font_height
            x1, y1 = x0 + mask.shape[1] - 1, y0 + mask.shape[0] - 1
            return x0, y0, x1, y1, glyph
    else:
        def get_sprite(self, row, col, char, fore, back):
            """Return a sprite for a given character."""
            if char not in self._glyphs:
                self._submit_char(char)
            mask = self._glyphs[char]
            glyph = [[(fore if _bit else back) for _bit in _row] for _row in mask]
            x0, y0 = (col-1) * self._mode.font_width, (row-1) * self._mode.font_height
            x1, y1 = x0 + len(mask[0]) - 1, y0 + len(mask) - 1
            return x0, y0, x1, y1, glyph



"""
PC-BASIC - pixels.py
Graphics buffer operations

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

try:
    import numpy
except ImportError:
    numpy = None

from ..base import error
from ..base import tokens as tk
from .. import values


class PixelBuffer(object):
    """Buffer for graphics on all screen pages."""

    def __init__(self, bwidth, bheight, bpages, bitsperpixel):
        """Initialise the graphics buffer to given pages and dimensions."""
        self.pages = [PixelPage(bwidth, bheight, num, bitsperpixel) for num in range(bpages)]
        self.width = bwidth
        self.height = bheight

    def copy_page(self, src, dst):
        """Copy source to destination page."""
        self.pages[dst].buffer[:] = self.pages[src].buffer[:]


class PixelPage(object):
    """Buffer for a screen page."""

    def __init__(self, bwidth, bheight, pagenum, bitsperpixel):
        """Initialise the screen buffer to given dimensions."""
        if numpy:
            self.buffer = numpy.zeros((bheight, bwidth), dtype=numpy.int8)
        else:
            self.buffer = [[0]*bwidth for _ in range(bheight)]
        self.width = bwidth
        self.height = bheight
        self.pagenum = pagenum
        self.bitsperpixel = bitsperpixel
        self.init_operations()

    def __getstate__(self):
        """Pickle the page."""
        pagedict = self.__dict__.copy()
        # lambdas can't be pickled
        pagedict['operations'] = None
        return pagedict

    def __setstate__(self, pagedict):
        """Initialise from pickled page."""
        self.__dict__.update(pagedict)
        self.init_operations()

    def put_pixel(self, x, y, attr):
        """Put a pixel in the buffer."""
        try:
            self.buffer[y][x] = attr
        except IndexError:
            pass

    def get_pixel(self, x, y):
        """Get attribute of a pixel in the buffer."""
        try:
            return self.buffer[y][x]
        except IndexError:
            return 0

    def fill_interval(self, x0, x1, y, attr):
        """Write a list of attributes to a scanline interval."""
        try:
            self.buffer[y][x0:x1+1] = [attr]*(x1-x0+1)
        except IndexError:
            pass

    if numpy:
        def init_operations(self):
            """Initialise operations closures."""
            self.operations = {
                tk.PSET: lambda x, y: x.__setitem__(slice(len(x)), y),
                tk.PRESET: lambda x, y: x.__setitem__(slice(len(x)), y.__xor__((1<<self.bitsperpixel) - 1)),
                tk.AND: lambda x, y: x.__iand__(y),
                tk.OR: lambda x, y: x.__ior__(y),
                tk.XOR: lambda x, y: x.__ixor__(y),
            }

        def put_interval(self, x, y, colours, mask=0xff):
            """Write a list of attributes to a scanline interval."""
            colours = numpy.array(colours).astype(int)
            inv_mask = 0xff ^ mask
            colours &= mask
            try:
                self.buffer[y, x:x+len(colours)] &= inv_mask
                self.buffer[y, x:x+len(colours)] |= colours
                return self.buffer[y, x:x+len(colours)]
            except IndexError:
                return numpy.zeros(len(colours), dtype=numpy.int8)

        def get_interval(self, x, y, length):
            """Return *view of* attributes of a scanline interval."""
            try:
                return self.buffer[y, x:x+length]
            except IndexError:
                return numpy.zeros(length, dtype=numpy.int8)

        def fill_rect(self, x0, y0, x1, y1, attr):
            """Apply solid attribute to an area."""
            if (x1 < x0) or (y1 < y0):
                return
            try:
                self.buffer[y0:y1+1, x0:x1+1].fill(attr)
            except IndexError:
                pass

        def put_rect(self, x0, y0, x1, y1, array, operation_token):
            """Apply numpy array [y][x] of attributes to an area."""
            if (x1 < x0) or (y1 < y0):
                return
            try:
                self.operations[operation_token](
                    self.buffer[y0:y1+1, x0:x1+1],
                    numpy.asarray(array)
                )
                return self.buffer[y0:y1+1, x0:x1+1]
            except IndexError:
                return numpy.zeros((y1-y0+1, x1-x0+1), dtype=numpy.int8)

        def get_rect(self, x0, y0, x1, y1):
            """Get *copy of* numpy array [y][x] of target area."""
            try:
                # our only user in module graphics needs a copy, so copy.
                return numpy.array(self.buffer[y0:y1+1, x0:x1+1])
            except IndexError:
                return numpy.zeros((y1-y0+1, x1-x0+1), dtype=numpy.int8)

        def move_rect(self, sx0, sy0, sx1, sy1, tx0, ty0):
            """Move pixels from an area to another, replacing with attribute 0."""
            w, h = sx1-sx0+1, sy1-sy0+1
            area = numpy.array(self.buffer[sy0:sy1+1, sx0:sx1+1])
            self.buffer[sy0:sy1+1, sx0:sx1+1] = numpy.zeros((h, w), dtype=numpy.int8)
            self.buffer[ty0:ty0+h, tx0:tx0+w] = area

        def get_until(self, x0, x1, y, c):
            """Get the attribute values of a scanline interval [x0, x1-1]."""
            if x0 == x1:
                return []
            toright = x1 > x0
            if not toright:
                x0, x1 = x1+1, x0+1
            try:
                arr = self.buffer[y, x0:x1]
            except IndexError:
                return []
            found = numpy.where(arr == c)
            if len(found[0]) > 0:
                if toright:
                    arr = arr[:found[0][0]]
                else:
                    arr = arr[found[0][-1]+1:]
            return list(arr.flatten())

    else:
        def init_operations(self):
            """Initialise operations closures."""
            self.operations = {
                tk.PSET: lambda x, y: y,
                tk.PRESET: lambda x, y: y ^ ((1<<self.bitsperpixel)-1),
                tk.AND: lambda x, y: x & y,
                tk.OR: lambda x, y: x | y,
                tk.XOR: lambda x, y: x ^ y,
            }

        def put_interval(self, x, y, colours, mask=0xff):
            """Write a list of attributes to a scanline interval."""
            inv_mask = 0xff ^ mask
            self.buffer[y][x:x+len(colours)] = [
                (c & mask) | (self.buffer[y][x+i] & inv_mask)
                for i, c in enumerate(colours)
            ]
            return self.buffer[y][x:x+len(colours)]

        def get_interval(self, x, y, length):
            """Return *view of* attributes of a scanline interval."""
            try:
                return self.buffer[y][x:x+length]
            except IndexError:
                return [0] * length

        def fill_rect(self, x0, y0, x1, y1, attr):
            """Apply solid attribute to an area."""
            if (x1 < x0) or (y1 < y0):
                return
            try:
                for y in range(y0, y1+1):
                    self.buffer[y][x0:x1+1] = [attr] * (x1-x0+1)
            except IndexError:
                pass

        def put_rect(self, x0, y0, x1, y1, array, operation_token):
            """Apply 2d list [y][x] of attributes to an area."""
            if (x1 < x0) or (y1 < y0):
                return
            try:
                for y in range(y0, y1+1):
                    self.buffer[y][x0:x1+1] = [
                        self.operations[operation_token](a, b)
                        for a, b in zip(self.buffer[y][x0:x1+1], array[y-y0])
                    ]
                return [self.buffer[y][x0:x1+1] for y in range(y0, y1+1)]
            except IndexError:
                return [[0]*(x1-x0+1) for _ in range(y1-y0+1)]

        def get_rect(self, x0, y0, x1, y1):
            """Get *copy of* 2d list [y][x] of target area."""
            try:
                return [self.buffer[y][x0:x1+1] for y in range(y0, y1+1)]
            except IndexError:
                return [[0]*(x1-x0+1) for _ in range(y1-y0+1)]

        def move_rect(self, sx0, sy0, sx1, sy1, tx0, ty0):
            """Move pixels from an area to another, replacing with attribute 0."""
            for y in range(0, sy1-sy0+1):
                row = self.buffer[sy0+y][sx0:sx1+1]
                self.buffer[sy0+y][sx0:sx1+1] = [0] * (sx1-sx0+1)
                self.buffer[ty0+y][tx0:tx0+(sx1-sx0+1)] = row

        def get_until(self, x0, x1, y, c):
            """Get the attribute values of a scanline interval [x0, x1-1]."""
            if x0 == x1:
                return []
            toright = x1 > x0
            if not toright:
                x0, x1 = x1+1, x0+1
            try:
                index = self.buffer[y][x0:x1].index(c)
            except ValueError:
                index = x1-x0
            return self.buffer[y][x0:x0+index]



"""
PC-BASIC - interface.video
Base class for video plugins

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import time

from ..compat import queue
from ..basic.base import signals


class VideoPlugin(object):
    """Base class for display/input interface plugins."""

    def __init__(self, input_queue, video_queue, **kwargs):
        """Setup the interface."""
        self.alive = True
        self.busy = False
        self._input_queue = input_queue
        self._video_queue = video_queue
        self._handlers = {
            signals.VIDEO_SET_MODE: self.set_mode,
            signals.VIDEO_PUT_GLYPH: self.put_glyph,
            signals.VIDEO_CLEAR_ROWS: self.clear_rows,
            signals.VIDEO_SCROLL_UP: self.scroll_up,
            signals.VIDEO_SCROLL_DOWN: self.scroll_down,
            signals.VIDEO_SET_PALETTE: self.set_palette,
            signals.VIDEO_SET_CURSOR_SHAPE: self.set_cursor_shape,
            signals.VIDEO_SET_CURSOR_ATTR: self.set_cursor_attr,
            signals.VIDEO_SHOW_CURSOR: self.show_cursor,
            signals.VIDEO_MOVE_CURSOR: self.move_cursor,
            signals.VIDEO_SET_PAGE: self.set_page,
            signals.VIDEO_COPY_PAGE: self.copy_page,
            signals.VIDEO_SET_BORDER_ATTR: self.set_border_attr,
            signals.VIDEO_BUILD_GLYPHS: self.build_glyphs,
            signals.VIDEO_PUT_PIXEL: self.put_pixel,
            signals.VIDEO_PUT_INTERVAL: self.put_interval,
            signals.VIDEO_FILL_INTERVAL: self.fill_interval,
            signals.VIDEO_PUT_RECT: self.put_rect,
            signals.VIDEO_FILL_RECT: self.fill_rect,
            signals.VIDEO_SET_CAPTION: self.set_caption_message,
            signals.VIDEO_SET_CLIPBOARD_TEXT: self.set_clipboard_text,
        }

    # called by Interface

    def cycle(self):
        """Video/input event cycle."""
        if self.alive:
            self._drain_queue()
        if self.alive:
            self._work()
            self._check_input()

    def sleep(self, ms):
        """Sleep a tick"""
        time.sleep(ms/1000.)

    # private methods

    def _drain_queue(self):
        """Drain signal queue."""
        while True:
            try:
                signal = self._video_queue.get(False)
            except queue.Empty:
                return True
            # putting task_done before the execution avoids hanging on join() after an exception
            self._video_queue.task_done()
            if signal.event_type == signals.VIDEO_QUIT:
                # close thread
                self.alive = False
            else:
                try:
                    self._handlers[signal.event_type](*signal.params)
                except KeyError:
                    pass

    # plugin overrides

    def __exit__(self, type, value, traceback):
        """Close the interface."""

    def __enter__(self):
        """Final initialisation."""
        return self

    def _work(self):
        """Display update cycle."""

    def _check_input(self):
        """Input devices update cycle."""

    # signal handlers

    def set_mode(self, mode_info):
        """Initialise a given text or graphics mode."""

    def set_caption_message(self, msg):
        """Add a message to the window caption."""

    def set_clipboard_text(self, text, mouse):
        """Put text on the clipboard."""

    def set_palette(self, rgb_palette_0, rgb_palette_1, pack_pixels):
        """Build the palette."""

    def set_border_attr(self, attr):
        """Change the border attribute."""

    def clear_rows(self, back_attr, start, stop):
        """Clear a range of screen rows."""

    def set_page(self, vpage, apage):
        """Set the visible and active page."""

    def copy_page(self, src, dst):
        """Copy source to destination page."""

    def show_cursor(self, cursor_on):
        """Change visibility of cursor."""

    def move_cursor(self, crow, ccol):
        """Move the cursor to a new position."""

    def set_cursor_attr(self, attr):
        """Change attribute of cursor."""

    def scroll_up(self, from_line, scroll_height, back_attr):
        """Scroll the screen up between from_line and scroll_height."""

    def scroll_down(self, from_line, scroll_height, back_attr):
        """Scroll the screen down between from_line and scroll_height."""

    def put_glyph(self, pagenum, row, col, char, is_fullwidth, fore, back, blink, underline):
        """Put a character at a given position."""

    def build_glyphs(self, new_dict):
        """Build a dict of glyphs for use in text mode."""

    def set_cursor_shape(self, width, height, from_line, to_line):
        """Build a sprite for the cursor."""

    def put_pixel(self, pagenum, x, y, index):
        """Put a pixel on the screen; callback to empty character buffer."""

    def fill_rect(self, pagenum, x0, y0, x1, y1, index):
        """Fill a rectangle in a solid attribute."""

    def fill_interval(self, pagenum, x0, x1, y, index):
        """Fill a scanline interval in a solid attribute."""

    def put_interval(self, pagenum, x, y, colours):
        """Write a list of attributes to a scanline interval."""

    def put_rect(self, pagenum, x0, y0, x1, y1, array):
        """Apply numpy array [y][x] of attribytes to an area."""



"""
PC-BASIC - compat.win32_console
Windows console support:
- unicode output for Python 2
- scroll prevention
- ANSI input & adjustable echo

(c) 2018--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.

Borrows code from colorama, which is Copyright Jonathan Hartley 2013. BSD 3-Clause license
"""

# pylint: disable=no-name-in-module, no-member

import os
import sys
import time
import msvcrt
import ctypes
from collections import deque
from ctypes import windll, wintypes, POINTER, byref, Structure, cast

from .base import PY2, wrap_input_stream, wrap_output_stream

if PY2:
    from .python2 import SimpleNamespace
else:
    from types import SimpleNamespace


# Windows virtual key codes, mapped to standard key names
KEYS = SimpleNamespace(
    PAGEUP = 0x21, # VK_PRIOR
    PAGEDOWN = 0x22, # VK_NEXT
    END = 0x23,
    HOME = 0x24,
    LEFT = 0x25,
    UP = 0x26,
    RIGHT = 0x27,
    DOWN = 0x28,
    INSERT = 0x2d,
    DELETE = 0x2e,
    F1 = 0x70,
    F2 = 0x71,
    F3 = 0x72,
    F4 = 0x73,
    F5 = 0x74,
    F6 = 0x75,
    F7 = 0x76,
    F8 = 0x77,
    F9 = 0x78,
    F10 = 0x79,
    F11 = 0x7a,
    F12 = 0x7b,
    #
    ALT = 0x12, # VK_MENU
)
VK_TO_KEY = {value: key for key, value in KEYS.__dict__.items()}
# alpha key codes
VK_TO_KEY.update({
    value: chr(value).lower() for value in range(0x30, 0x5b)
})

# control key state bit flags
# CAPSLOCK_ON = 0x0080,
# ENHANCED_KEY = 0x0100,
# LEFT_ALT_PRESSED = 0x0002,
# LEFT_CTRL_PRESSED = 0x0008,
# NUMLOCK_ON = 0x0020,
# RIGHT_ALT_PRESSED = 0x0001,
# RIGHT_CTRL_PRESSED = 0x0004,
# SCROLLLOCK_ON = 0x0040,
# SHIFT_PRESSED = 0x0010,
MODS = dict(
    CTRL = 0x0c,
    ALT = 0x03,
    SHIFT = 0x10,
)

# character attributes, from wincon.h
NORMAL = 0x00
BRIGHT = 0x08


##############################################################################
# ctypes wrappers

class KEY_EVENT_RECORD(Structure):
    _fields_ = (
        ('bKeyDown', wintypes.BOOL), #32 bit?
        ('wRepeatCount', wintypes.WORD), #16
        ('wVirtualKeyCode', wintypes.WORD),#16
        ('wVirtualScanCode', wintypes.WORD), #16
        # union with CHAR AsciiChar
        ('UnicodeChar', wintypes.WCHAR), #32
        ('dwControlKeyState', wintypes.DWORD), #32
        # note that structure is in a union with other event records
        # but it is the largest type. mouseeventrecord is 128 bytes
    )

class INPUT_RECORD(Structure):
    _fields_ = (
        ('EventType', wintypes.WORD),
        # union of many event types but we only care about key events
        # total size is 16 bytes
        ('KeyEvent', KEY_EVENT_RECORD),
    )

class CHAR_INFO(Structure):
    _fields_ = (
        ('UnicodeChar', wintypes.WCHAR),
        ('Attributes', wintypes.WORD),
    )

class CONSOLE_SCREEN_BUFFER_INFO(Structure):
    _fields_ = (
        ('dwSize', wintypes._COORD),
        ('dwCursorPosition', wintypes._COORD),
        ('wAttributes', wintypes.WORD),
        ('srWindow', wintypes.SMALL_RECT),
        ('dwMaximumWindowSize', wintypes._COORD),
    )

class CONSOLE_CURSOR_INFO(Structure):
    _fields_ = (
        ("dwSize", wintypes.DWORD),
        ("bVisible", wintypes.BOOL),
    )

class CONSOLE_SCREEN_BUFFER_INFOEX(Structure):
    _fields_ = (
        ('cbSize', wintypes.ULONG),
        ('dwSize', wintypes._COORD),
        ('dwCursorPosition', wintypes._COORD),
        ('wAttributes', wintypes.WORD),
        ('srWindow', wintypes.SMALL_RECT),
        ('dwMaximumWindowSize', wintypes._COORD),
        ('wPopupAttributes', wintypes.WORD),
        ('bFullscreenSupported', wintypes.BOOL),
        ('ColorTable', wintypes.DWORD*16),
    )

_GetStdHandle = windll.kernel32.GetStdHandle
_GetStdHandle.argtypes = (wintypes.DWORD,)
_GetStdHandle.restype = wintypes.HANDLE

_WriteConsoleW = windll.kernel32.WriteConsoleW
_WriteConsoleW.argtypes = (
    wintypes.HANDLE, wintypes.LPCWSTR, wintypes.DWORD, POINTER(wintypes.DWORD), wintypes.LPVOID
)

_WriteConsoleOutputW = windll.kernel32.WriteConsoleOutputW
_WriteConsoleOutputW.argtypes = (
    wintypes.HANDLE, POINTER(CHAR_INFO),
    wintypes._COORD, wintypes._COORD, POINTER(wintypes.SMALL_RECT)
)

_ReadConsoleInputW = windll.kernel32.ReadConsoleInputW
_ReadConsoleInputW.argtypes = (
    wintypes.HANDLE, POINTER(INPUT_RECORD), wintypes.DWORD, POINTER(wintypes.DWORD)
)

_GetNumberOfConsoleInputEvents = windll.kernel32.GetNumberOfConsoleInputEvents
_GetNumberOfConsoleInputEvents.argtypes = (wintypes.HANDLE, POINTER(wintypes.DWORD))

_GetConsoleScreenBufferInfo = windll.kernel32.GetConsoleScreenBufferInfo
_GetConsoleScreenBufferInfo.argtypes = (wintypes.HANDLE, POINTER(CONSOLE_SCREEN_BUFFER_INFO))

_SetConsoleTitleW = windll.kernel32.SetConsoleTitleW
_SetConsoleTitleW.argtypes = (wintypes.LPCWSTR,)
_SetConsoleTitleW.restype = wintypes.BOOL

_FillConsoleOutputCharacterW = windll.kernel32.FillConsoleOutputCharacterW
_FillConsoleOutputCharacterW.argtypes = (
    wintypes.HANDLE,
    wintypes.WCHAR,
    wintypes.DWORD,
    wintypes._COORD,
    POINTER(wintypes.DWORD),
)
_FillConsoleOutputCharacterW.restype = wintypes.BOOL

_FillConsoleOutputAttribute = windll.kernel32.FillConsoleOutputAttribute
_FillConsoleOutputAttribute.argtypes = (
    wintypes.HANDLE,
    wintypes.WORD,
    wintypes.DWORD,
    wintypes._COORD,
    POINTER(wintypes.DWORD),
)
_FillConsoleOutputAttribute.restype = wintypes.BOOL

_SetConsoleCursorPosition = windll.kernel32.SetConsoleCursorPosition
_SetConsoleCursorPosition.argtypes = (wintypes.HANDLE, wintypes._COORD)
_SetConsoleCursorPosition.restype = wintypes.BOOL

_GetConsoleCursorInfo = windll.kernel32.GetConsoleCursorInfo
_GetConsoleCursorInfo.argtypes = (wintypes.HANDLE, POINTER(CONSOLE_CURSOR_INFO))

_SetConsoleCursorInfo = windll.kernel32.SetConsoleCursorInfo
_SetConsoleCursorInfo.argtypes = (wintypes.HANDLE, POINTER(CONSOLE_CURSOR_INFO))

_SetConsoleScreenBufferInfoEx = windll.kernel32.SetConsoleScreenBufferInfoEx
_SetConsoleScreenBufferInfoEx.argtypes = (wintypes.HANDLE, POINTER(CONSOLE_SCREEN_BUFFER_INFOEX))

_GetConsoleScreenBufferInfoEx = windll.kernel32.GetConsoleScreenBufferInfoEx
_GetConsoleScreenBufferInfoEx.argtypes = (wintypes.HANDLE, POINTER(CONSOLE_SCREEN_BUFFER_INFOEX))

_ScrollConsoleScreenBuffer = windll.kernel32.ScrollConsoleScreenBufferW
_ScrollConsoleScreenBuffer.argtypes = (
    wintypes.HANDLE,
    POINTER(wintypes.SMALL_RECT),
    POINTER(wintypes.SMALL_RECT),
    wintypes._COORD,
    POINTER(CHAR_INFO),
)

_SetConsoleTextAttribute = windll.kernel32.SetConsoleTextAttribute
_SetConsoleTextAttribute.argtypes = (wintypes.HANDLE, wintypes.WORD)
_SetConsoleTextAttribute.restype = wintypes.BOOL

_SetConsoleScreenBufferSize = windll.kernel32.SetConsoleScreenBufferSize
_SetConsoleScreenBufferSize.argtypes = (wintypes.HANDLE, wintypes._COORD)

_SetConsoleWindowInfo = windll.kernel32.SetConsoleWindowInfo
_SetConsoleWindowInfo.argtypes = (
    wintypes.HANDLE,
    wintypes.BOOL,
    POINTER(wintypes.SMALL_RECT),
)

_SetConsoleMode = windll.kernel32.SetConsoleMode
_SetConsoleMode.argtypes = (
    wintypes.HANDLE,
    wintypes.DWORD,
)

_GetConsoleMode = windll.kernel32.GetConsoleMode
_GetConsoleMode.argtypes = (
    wintypes.HANDLE,
    POINTER(wintypes.DWORD),
)
_GetConsoleMode.restype = wintypes.BOOL


PHANDLER_ROUTINE = ctypes.WINFUNCTYPE(wintypes.BOOL, wintypes.DWORD)
_SetConsoleCtrlHandler = windll.kernel32.SetConsoleCtrlHandler
_SetConsoleCtrlHandler.argtypes = (
    PHANDLER_ROUTINE,
    wintypes.BOOL,
)
_SetConsoleCtrlHandler.restype = wintypes.BOOL


def GetConsoleScreenBufferInfo(handle):
    csbi = CONSOLE_SCREEN_BUFFER_INFO()
    _GetConsoleScreenBufferInfo(handle, byref(csbi))
    return csbi

def GetConsoleScreenBufferInfoEx(handle):
    csbie = CONSOLE_SCREEN_BUFFER_INFOEX()
    csbie.cbSize = wintypes.ULONG(ctypes.sizeof(csbie))
    _GetConsoleScreenBufferInfoEx(handle, byref(csbie))
    # work around Windows bug
    # see https://stackoverflow.com/questions/35901572/setconsolescreenbufferinfoex-bug
    csbie.srWindow.Bottom += 1
    csbie.srWindow.Right += 1
    return csbie

def FillConsoleOutputCharacter(handle, char, length, start):
    length = wintypes.DWORD(length)
    num_written = wintypes.DWORD(0)
    _FillConsoleOutputCharacterW(handle, char, length, start, byref(num_written))
    return num_written.value

def FillConsoleOutputAttribute(handle, attr, length, start):
    attribute = wintypes.WORD(attr)
    length = wintypes.DWORD(length)
    return _FillConsoleOutputAttribute(handle, attribute, length, start, byref(wintypes.DWORD()))

def GetConsoleCursorInfo(handle):
    cci = CONSOLE_CURSOR_INFO()
    _GetConsoleCursorInfo(handle, byref(cci))
    return cci

def SetConsoleCursorInfo(handle, cci):
    _SetConsoleCursorInfo(handle, byref(cci))

def ScrollConsoleScreenBuffer(handle, scroll_rect, clip_rect, new_position, char, attr):
    char_info = CHAR_INFO(char, attr)
    _ScrollConsoleScreenBuffer(
        handle, byref(scroll_rect), byref(clip_rect), new_position, byref(char_info)
    )

def GetConsoleMode(handle):
    mode = wintypes.DWORD()
    _GetConsoleMode(handle, mode)
    return mode


HSTDIN = _GetStdHandle(-10)
HSTDOUT = _GetStdHandle(-11)
HSTDERR = _GetStdHandle(-12)


def _write_console(handle, unistr):
    """Write character to console, avoid scroll on bottom line."""
    csbi = CONSOLE_SCREEN_BUFFER_INFO()
    _GetConsoleScreenBufferInfo(handle, byref(csbi))
    col, row = csbi.dwCursorPosition.X, csbi.dwCursorPosition.Y
    width, height = csbi.dwSize.X, csbi.dwSize.Y
    for ch in unistr:
        if (row == height-1 and col >= width - 1 and ch != u'\n'):
            ci = CHAR_INFO(ch, csbi.wAttributes)
            # do not advance cursor if we're on the last position of the
            # screen buffer, to avoid unwanted scrolling.
            _WriteConsoleOutputW(
                handle, byref(ci), wintypes._COORD(1, 1), wintypes._COORD(0, 0),
                wintypes.SMALL_RECT(col, row, col, row)
            )
        else:
            _WriteConsoleW(
                handle, ch, 1,
                byref(wintypes.DWORD()), byref(wintypes.DWORD())
            )


##############################################################################
# console class

@PHANDLER_ROUTINE
def _ctrl_handler(fdwCtrlType):
    """Handle Ctrl-Break event."""
    # CTRL_BREAK_EVENT
    return (fdwCtrlType == 1)


class Win32Console(object):
    """Win32API-based console implementation."""

    def __init__(self):
        """Set up console"""
        # preserve original settings
        self._orig_csbie = GetConsoleScreenBufferInfoEx(HSTDOUT)
        self._orig_stdin_mode = GetConsoleMode(HSTDIN)
        self._attrs = self._orig_csbie.wAttributes
        # input
        self._input_buffer = deque()
        self._echo = True

    def set_raw(self):
        """Enter raw terminal mode (no echo, don't exit on ctrl-C)."""
        self._echo = False
        # unset ENABLE_PROCESSED_INPUT
        _SetConsoleMode(HSTDIN, wintypes.DWORD(self._orig_stdin_mode.value & ~0x0001))
        # don't exit on ctrl-Break
        _SetConsoleCtrlHandler(_ctrl_handler, True)

    def unset_raw(self):
        """Leave raw terminal mode."""
        self._echo = True
        _SetConsoleMode(HSTDOUT, self._orig_stdin_mode)

    def key_pressed(self):
        """key pressed on keyboard."""
        return msvcrt.kbhit()

    def set_caption(self, caption):
        """Set terminal caption."""
        _SetConsoleTitleW(caption)

    def resize(self, height, width):
        """Resize terminal."""
        csbi = GetConsoleScreenBufferInfo(HSTDOUT)
        # SetConsoleScreenBufferSize can't make the buffer smaller than the window
        # SetConsoleWindowInfo can't make the window larger than the buffer (in either direction)
        # allow for both shrinking and growing by calling one of them twice,
        # for each direction separately
        new_size = wintypes._COORD(width, csbi.dwSize.Y)
        new_window = wintypes.SMALL_RECT(0, 0, width-1, csbi.dwSize.Y-1)
        _SetConsoleScreenBufferSize(HSTDOUT, new_size)
        _SetConsoleWindowInfo(HSTDOUT, True, new_window)
        _SetConsoleScreenBufferSize(HSTDOUT, new_size)
        new_size = wintypes._COORD(width, height)
        new_window = wintypes.SMALL_RECT(0, 0, width-1, height-1)
        _SetConsoleScreenBufferSize(HSTDOUT, new_size)
        _SetConsoleWindowInfo(HSTDOUT, True, new_window)
        _SetConsoleScreenBufferSize(HSTDOUT, new_size)

    ##########################################################################
    # output

    def write(self, unistr):
        """Write text to the console."""
        _write_console(HSTDOUT, unistr)

    def clear(self):
        """Clear the screen."""
        csbi = GetConsoleScreenBufferInfo(HSTDOUT)
        # fill the entire screen with blanks
        FillConsoleOutputCharacter(
            HSTDOUT, u' ', csbi.dwSize.X * csbi.dwSize.Y, wintypes._COORD(0, 0)
        )
        # now set the buffer's attributes accordingly
        FillConsoleOutputAttribute(
            HSTDOUT, self._attrs, csbi.dwSize.X * csbi.dwSize.Y, wintypes._COORD(0, 0)
        )
        _SetConsoleCursorPosition(HSTDOUT, wintypes._COORD(0, 0))

    def clear_row(self):
        """Clear the current row."""
        csbi = GetConsoleScreenBufferInfo(HSTDOUT)
        from_coord = wintypes._COORD(0, csbi.dwCursorPosition.Y)
        # fill the entire screen with blanks
        FillConsoleOutputCharacter(HSTDOUT, u' ', csbi.dwSize.X, from_coord)
        # now set the buffer's attributes accordingly
        FillConsoleOutputAttribute(HSTDOUT, self._attrs, csbi.dwSize.X, from_coord)

    def set_cursor_colour(self, colour):
        """Set the current cursor colour attribute - not supported."""

    def show_cursor(self, block=False):
        """Show the cursor."""
        curs_info = GetConsoleCursorInfo(HSTDOUT)
        curs_info.bVisible = True
        curs_info.dwSize = 100 if block else 20
        SetConsoleCursorInfo(HSTDOUT, curs_info)

    def hide_cursor(self):
        """Hide the cursor."""
        curs_info = GetConsoleCursorInfo(HSTDOUT)
        curs_info.bVisible = False
        SetConsoleCursorInfo(HSTDOUT, curs_info)

    def move_cursor_left(self, n):
        """Move cursor n cells to the left."""
        self._move_cursor(0, -n)

    def move_cursor_right(self, n):
        """Move cursor n cells to the right."""
        self._move_cursor(0, n)

    def _move_cursor(self, rows, cols):
        """Move cursor relative to current position."""
        csbi = GetConsoleScreenBufferInfo(HSTDOUT)
        position = csbi.dwCursorPosition
        self.move_cursor_to(position.Y+1 + rows, position.X+1 + cols)

    def move_cursor_to(self, row, col):
        """Move cursor to a new position (1,1 is top left)."""
        csbi = GetConsoleScreenBufferInfo(HSTDOUT)
        row, col = row-1, col-1
        while col >= csbi.dwSize.X:
            col -= csbi.dwSize.X
            row += 1
        while col < 0:
            col += csbi.dwSize.X
            row -= 1
        # If the position is out of range, do nothing.
        if row >= 0 and col >= 0:
            _SetConsoleCursorPosition(HSTDOUT, wintypes._COORD(col, row))

    def _scroll(self, start, stop, rows):
        if not rows:
            return
        csbi = GetConsoleScreenBufferInfo(HSTDOUT)
        # absolute position of window in screen buffer
        # interpret other coordinates as relative to the window
        window = csbi.srWindow
        # scroll region
        clip_rect = wintypes.SMALL_RECT(
            window.Left, window.Top + start, window.Right, window.Top + stop
        )
        if rows > 0:
            region = wintypes.SMALL_RECT(window.Left, window.Top + rows, window.Right, window.Bottom)
            new_pos = wintypes._COORD(window.Left, window.Top)
        else:
            # minus signs since rows is a negative number
            region = wintypes.SMALL_RECT(window.Left, window.Top, window.Right, window.Bottom - rows)
            new_pos = wintypes._COORD(window.Left, window.Top - rows)
        # workaround: in this particular case, Windows doesn't seem to respect the clip area.
        if (
                clip_rect.Bottom == window.Bottom-1 and
                region.Bottom >= window.Bottom-1 and
                new_pos.Y < region.Top
            ):
            # first scroll everything up
            clip_rect.Bottom = window.Bottom
            bottom, region.Bottom = region.Bottom, window.Bottom
            ScrollConsoleScreenBuffer(HSTDOUT, region, clip_rect, new_pos, u' ', self._attrs)
            # and then scroll the bottom back down
            new_pos.Y = window.Bottom
            region.Top = bottom-1
            ScrollConsoleScreenBuffer(HSTDOUT, region, clip_rect, new_pos, u' ', self._attrs)
        else:
            ScrollConsoleScreenBuffer(HSTDOUT, region, clip_rect, new_pos, u' ', self._attrs)

    def scroll_up(self, top, bottom):
        """Scroll the region between top and bottom one row up."""
        self._scroll(top-1, bottom-1, 1)

    def scroll_down(self, top, bottom):
        """Scroll the region between top and bottom one row down."""
        self._scroll(top-1, bottom-1, -1)

    def reset(self):
        """Reset to default attributes."""
        _SetConsoleScreenBufferInfoEx(HSTDOUT, byref(self._orig_csbie))
        self._attrs = self._orig_csbie.wAttributes
        self.show_cursor()

    def set_attributes(self, fore, back, blink, underline):
        """Set current attributes."""
        self._attrs = fore % 8 + back * 16 + (BRIGHT if fore > 8 else NORMAL)
        _SetConsoleTextAttribute(HSTDOUT, self._attrs)

    def set_palette_entry(self, attr, red, green, blue):
        """Set palette entry for attribute (0--16)."""
        csbie = GetConsoleScreenBufferInfoEx(HSTDOUT)
        csbie.ColorTable[attr] = (
            0x00010000 * blue + 0x00000100 * green + 0x00000001 * red
        )
        _SetConsoleScreenBufferInfoEx(HSTDOUT, byref(csbie))

    ##########################################################################
    # input

    def read_key(self):
        """
        Read keypress from console. Non-blocking.
        Returns tuple (unicode, keycode, set of mods)
        """
        self._fill_buffer(blocking=False)
        if not self._input_buffer:
            return (u'', None, {})
        return self._input_buffer.popleft()

    def read_all_chars(self):
        """Read all characters in the buffer."""
        self._fill_buffer(blocking=False)
        closed = False
        if self._input_buffer and self._input_buffer[-1][0] == u'\x1a':
            closed = True
            self._input_buffer.pop()
            if not self._input_buffer:
                return None
        output = u''.join(
            _char for _char in self._input_buffer
        )
        self._input_buffer.clear()
        if closed:
            self._input_buffer.append(u'\x1a')
        return output

    def _fill_buffer(self, blocking):
        """Interpret all key events."""
        while True:
            nevents = wintypes.DWORD()
            _GetNumberOfConsoleInputEvents(HSTDIN, byref(nevents))
            if not nevents.value and not blocking:
                return
            # only ever block on first loop
            blocking = False
            if nevents.value > 0:
                input_buffer = (INPUT_RECORD * nevents.value)()
                nread = wintypes.DWORD()
                _ReadConsoleInputW(
                    HSTDIN,
                    cast(input_buffer, POINTER(INPUT_RECORD)),
                    nevents.value, byref(nread)
                )
                for event in input_buffer:
                    if event.EventType != 1: # KEY_EVENT
                        continue
                    char, key, mods = self._translate_event(event)
                    if char or key:
                        self._input_buffer.append((char, key, mods))
                        if char == u'\x1a':
                            # ctrl-z is end of input on windows console
                            return
                        if self._echo:
                            _write_console(HSTDOUT, char.replace(u'\r', u'\n'))
            time.sleep(0.01)

    def _translate_event(self, event):
        char = event.KeyEvent.UnicodeChar
        key = event.KeyEvent.wVirtualKeyCode
        control = event.KeyEvent.dwControlKeyState
        if char == u'\0':
            # windows uses null-terminated strings so \0 means no output
            char = u''
        if not event.KeyEvent.bKeyDown:
            # key-up event for unicode Alt+HEX input
            if event.KeyEvent.wVirtualKeyCode == KEYS.ALT:
                return char, None, set()
            # ignore other key-up events
            return u'', None, set()
        # decode modifier bit flags
        mods = set(key for key, mask in MODS.items() if control & mask)
        key = VK_TO_KEY.get(key, None)
        return char, key, mods


def _has_console():
    """Determine if we have a console attached or are a GUI app."""
    try:
        return bool(_GetConsoleMode(HSTDOUT, byref(wintypes.DWORD())))
    except Exception as e:
        return False


IS_CONSOLE_APP = _has_console()

if _has_console():
    console = Win32Console()
else:
    console = None


##############################################################################
# non-blocking input

def read_all_available(stream):
    """Read all available characters from a stream; nonblocking; None if closed."""
    # are we're reading from (wrapped) stdin or not?
    if hasattr(stream, 'isatty') and stream.isatty():
        # this is shaky - try to identify unicode vs bytes stream
        is_unicode_stream = hasattr(stream, 'buffer')
        unistr = console.read_all_chars()
        if is_unicode_stream or unistr is None:
            return unistr
        else:
            return unistr.encode(stdin.encoding, 'replace')
    else:
        # this would work on unix too
        # just read the whole file and be done with it
        return stream.read() or None


##############################################################################
# standard i/o

if PY2:

    class _StreamWrapper(object):
        """Delegating stream wrapper."""

        def __init__(self, stream, handle, encoding='utf-8'):
            self._wrapped = stream
            self._handle = handle
            self.encoding = encoding

        def __getattr__(self, attr):
            return getattr(self._wrapped, attr)


    class _ConsoleOutput(_StreamWrapper):
        """Bytes stream wrapper using Unicode API, to replace Python2 sys.stdout."""

        def write(self, bytestr):
            if not isinstance(bytestr, bytes):
                raise TypeError('write() argument must be bytes, not %s' % type(bytestr))
            unistr = bytestr.decode(self.encoding)
            _write_console(self._handle, unistr)


    class _ConsoleInput(_StreamWrapper):
        """Bytes stream wrapper using Unicode API, to replace Python2 sys.stdin."""

        def __init__(self, encoding='utf-8'):
            _StreamWrapper.__init__(self, sys.stdin, HSTDIN, encoding)

        def read(self, size=-1):
            output = bytearray()
            while size < 0 or len(output) < size:
                key = console.read_key()
                if isinstance(key, int):
                    continue
                output.append(key.encode(self.encoding))
            return bytes(output)


    if sys.stdin.isatty():
        stdin = wrap_input_stream(_ConsoleInput())
    else:
        stdin = wrap_input_stream(sys.stdin)

    if sys.stdout.isatty():
        stdout = wrap_output_stream(_ConsoleOutput(sys.stdout, HSTDOUT))
    else:
        stdout = wrap_output_stream(sys.stdout)

    if sys.stderr.isatty():
        stderr = wrap_output_stream(_ConsoleOutput(sys.stderr, HSTDERR))
    else:
        stderr = wrap_output_stream(sys.stderr)

else:
    stdin = sys.stdin
    stdout = sys.stdout
    stderr = sys.stderr


# set stdio as binary, to avoid Windows messing around with CRLFs
# only do this for redirected output, as it breaks interactive Python sessions
try:
    if not sys.stdin.isatty():
        msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
    if not sys.stdout.isatty():
        msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)
except EnvironmentError:
    # raises an error if started in gui mode, as we have no stdio
    pass



"""
PC-BASIC - compat.base
Cross-platform compatibility utilities

(c) 2018--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import re
import contextlib
import sys
import platform
import codecs


# Python major version
PY2 = sys.version_info.major == 2

# platform constants
WIN32 = sys.platform == 'win32'
MACOS = sys.platform == 'darwin'

# 64-bit (needed for Windows binary modules)
X64 = platform.architecture()[0] == '64bit'

# platform tag for libraries
if WIN32:
    PLATFORM = sys.platform + ('_x64' if X64 else '_x86')
else:
    PLATFORM = sys.platform

# user configuration and state directories
HOME_DIR = os.path.expanduser(u'~')

if WIN32:
    USER_CONFIG_HOME = os.getenv(u'APPDATA')
    USER_DATA_HOME = USER_CONFIG_HOME
elif MACOS:
    USER_CONFIG_HOME = os.path.join(HOME_DIR, u'Library', u'Application Support')
    USER_DATA_HOME = USER_CONFIG_HOME
else:
    USER_CONFIG_HOME = os.environ.get(u'XDG_CONFIG_HOME') or os.path.join(HOME_DIR, u'.config')
    USER_DATA_HOME = os.environ.get(u'XDG_DATA_HOME') or os.path.join(HOME_DIR, u'.local', u'share')

# package/executable directory
if hasattr(sys, 'frozen'):
    # we're a package: get the directory of the packaged executable
    # (__file__ is undefined in frozen packages)
    # this is for cx_Freeze's package layout
    BASE_DIR = os.path.join(os.path.dirname(sys.executable), 'lib', 'pcbasic')
else:
    BASE_DIR = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))


# unicode stream wrappers

def wrap_output_stream(stream):
    """Wrap std bytes streams to make them behave more like in Python 3."""
    wrapped = codecs.getwriter(stream.encoding or 'utf-8')(stream)
    wrapped.buffer = stream
    return wrapped

def wrap_input_stream(stream):
    """Wrap std bytes streams to make them behave more like in Python 3."""
    wrapped = codecs.getreader(stream.encoding or 'utf-8')(stream)
    wrapped.buffer = stream
    return wrapped


# utility functions, this has to go somewhere...

def split_quoted(line, split_by=u'\s', quote=u'"', strip_quotes=False):
    """Split by separators, preserving quoted blocks."""
    chunks = re.findall(u'[^%s%s][^%s]*|%s.+?%s' % (quote, split_by, split_by, quote, quote), line)
    if strip_quotes:
        chunks = [c.strip(quote) for c in chunks]
    return chunks

@contextlib.contextmanager
def muffle(std_stream):
    """Suppress stdout or stderr messages."""
    save = None
    try:
        try:
            # save the file descriptor for the target stream
            save = os.dup(std_stream.fileno())
        except EnvironmentError:
            yield
            return
        # http://stackoverflow.com/questions/977840/
        # redirecting-fortran-called-via-f2py-output-in-python/978264#978264
        with open(os.devnull, 'w') as null:
            # put /dev/null fds on 1 (stdout) or 2 (stderr)
            os.dup2(null.fileno(), std_stream.fileno())
            # do stuff
            try:
                yield
            finally:
                std_stream.flush()
                # restore file descriptors
                os.dup2(save, std_stream.fileno())
    finally:
        if save is not None:
            os.close(save)




NAME = u'PC-BASIC'
PACKAGE = 'pcbasic'
VERSION = u'2.0.2'
COPYRIGHT = u'(C) Copyright 2013--2019 Rob Hagemans.'
LICENCE = u'GPLv3'
AUTHOR = u'Rob Hagemans'
EMAIL = u'robhagemans@yahoo.co.uk'
URL = u'http://pc-basic.org'
DESCRIPTION = u'A free, cross-platform emulator for the GW-BASIC family of interpreters.'
LONG_DESCRIPTION = u'''\
PC-BASIC is a free, cross-platform interpreter for GW-BASIC, Advanced BASIC (BASICA), PCjr Cartridge Basic and Tandy 1000 GWBASIC.
It interprets these BASIC dialects with a high degree of accuracy, aiming for bug-for-bug compatibility.
PC-BASIC emulates the most common video and audio hardware on which these BASICs used to run.
PC-BASIC runs plain-text, tokenised and protected .BAS files.
It implements floating-point arithmetic in the Microsoft Binary Format (MBF) and can therefore
read and write binary data files created by GW-BASIC.'''
CLASSIFIERS = [
    u'Development Status :: 4 - Beta',
    u'Intended Audience :: End Users/Desktop',
    u'Intended Audience :: Developers',
    u'Topic :: System :: Emulators',
    u'Topic :: Software Development :: Interpreters',
    u'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    u'Programming Language :: Python :: 2.7',
    u'Programming Language :: Python :: 3.5',
]
KEYWORDS = [u'emulator', u'interpreter', u'basic', u'retro', u'legacy', u'gwbasic', u'basica', u'pcjr', u'tandy']



"""
PC-BASIC - data package
HEX font loader

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import sys
import pkg_resources
import logging
import binascii

from ..compat import iteritems, itervalues, unichr

from ..basic.codepage import PRINTABLE_ASCII
from .resources import get_data, ResourceFailed


FONT_DIR = u'fonts'
FONT_PATTERN = u'{path}/{name}_{height:02d}.hex'
FONTS = sorted(
    set(name.split(u'_', 1)[0]
    for name in pkg_resources.resource_listdir(__name__, FONT_DIR)
    if name.lower().endswith(u'.hex'))
)


def read_fonts(codepage_dict, font_families, warn=False):
    """Load font typefaces."""
    # load the graphics fonts, including the 8-pixel RAM font
    # use set() for speed - lookup is O(1) rather than O(n) for list
    unicode_needed = set(itervalues(codepage_dict))
    # break up any grapheme clusters and add components to set of needed glyphs
    unicode_needed |= set(c for cluster in unicode_needed if len(cluster) > 1 for c in cluster)
    # substitutes is in reverse order: { yen: backslash }
    substitutes = {
        grapheme_cluster: unichr(ord(cp_point))
        for cp_point, grapheme_cluster in iteritems(codepage_dict)
        if (
            cp_point in PRINTABLE_ASCII
            and (len(grapheme_cluster) > 1
            or ord(grapheme_cluster) != ord(cp_point))
        )
    }
    fonts = {}
    # load fonts, height-16 first
    for height in (16, 14, 8):
        # load a Unifont .hex font and take the codepage subset
        font_files = []
        for name in font_families:
            try:
                font_files.append(
                    get_data(FONT_PATTERN, path=FONT_DIR, name=name, height=height)
                )
            except ResourceFailed as e:
                if warn:
                    logging.debug(e)
        fonts[height] = FontLoader(height).load_hex(
            font_files, unicode_needed, substitutes, warn=warn
        )
        # fix missing code points font based on 16-line font
        if fonts[16]:
            fonts[height].fix_missing(unicode_needed, fonts[16])
    if 8 in fonts:
        fonts[9] = fonts[8]
    # convert keys from unicode to codepage
    fonts = {
        height: {
            c: font._fontdict[uc]
            for c, uc in iteritems(codepage_dict) if uc in font._fontdict
        }
        for height, font in iteritems(fonts)
    }
    return {height: font for height, font in iteritems(fonts)}


class FontLoader(object):
    """Single-height bitfont."""

    def __init__(self, height):
        """Initialise the font."""
        self._height = height
        self._fontdict = {}

    def load_hex(self, hex_resources, unicode_needed, substitutes, warn=True):
        """Load a set of overlaying unifont .hex files."""
        self._fontdict = {}
        all_needed = unicode_needed | set(substitutes)
        for hexres in reversed(hex_resources):
            if hexres is None:
                continue
            for line in hexres.splitlines():
                # ignore empty lines and comment lines (first char is #)
                if (not line) or (line[:1] == b'#'):
                    continue
                # strip off comments
                # split unicodepoint and hex string (max 32 chars)
                ucs_str, fonthex = line.split(b'#')[0].split(b':')
                ucs_sequence = ucs_str.split(b',')
                fonthex = fonthex.strip()
                # extract codepoint and hex string;
                # discard anything following whitespace; ignore malformed lines
                try:
                    # construct grapheme cluster
                    c = u''.join(unichr(int(ucshex.strip(), 16)) for ucshex in ucs_sequence)
                    # skip grapheme clusters we won't need
                    if c not in all_needed:
                        continue
                    # skip chars we already have
                    if (c in self._fontdict):
                        continue
                    # string must be 32-byte or 16-byte; cut to required font size
                    if len(fonthex) < 32:
                        raise ValueError
                    if len(fonthex) < 64:
                        fonthex = fonthex[:2*self._height]
                    else:
                        fonthex = fonthex[:4*self._height]
                    self._fontdict[c] = binascii.unhexlify(fonthex)
                except Exception as e:
                    logging.warning('Could not parse line in font file: %s', repr(line))
        # substitute code points
        self._fontdict.update({
            old: self._fontdict[new]
            for (new, old) in iteritems(substitutes)
            if new in self._fontdict
        })
        # char 0 should always be defined and empty
        self._fontdict[u'\0'] = b'\0' * self._height
        self._combine_glyphs(unicode_needed)
        # in debug mode, check if we have all needed glyphs
        if warn:
            self._warn_missing(unicode_needed)
        return self

    def _combine_glyphs(self, unicode_needed):
        """Fix missing grapheme clusters by combining components."""
        for cluster in unicode_needed:
            if cluster not in self._fontdict:
                # try to combine grapheme clusters first
                if len(cluster) > 1:
                    # combine strings
                    clusterglyph = bytearray(self._height)
                    try:
                        for c in cluster:
                            for y, row in enumerate(self._fontdict[c]):
                                clusterglyph[y] |= ord(row)
                    except KeyError as e:
                        logging.debug(
                            'Could not combine grapheme cluster %s, missing %r [%s]', cluster, c, c
                        )
                    self._fontdict[cluster] = bytes(clusterglyph)

    def _warn_missing(self, unicode_needed, max_warnings=3):
        """Check if we have all needed glyphs."""
        # fontdict: unicode char -> glyph
        missing = unicode_needed - set(self._fontdict)
        warnings = 0
        for u in missing:
            warnings += 1
            logging.debug('Code point u+%x not represented in font', ord(u))
            if warnings == max_warnings:
                logging.debug('Further code point warnings suppressed.')
                break

    def fix_missing(self, unicode_needed, font16):
        """Fill in missing codepoints in font using 16-line font or blanks."""
        if self._height == 16:
            return
        for c in unicode_needed:
            if c not in self._fontdict:
                # try to construct from 16-bit font
                try:
                    s16 = list(font16.fontdict[c])
                    start = (16 - self._height) // 2
                    if len(s16) == 16:
                        self._fontdict[c] = b''.join([s16[i] for i in range(start, 16-start)])
                    else:
                        self._fontdict[c] = b''.join([s16[i] for i in range(start*2, 32-start*2)])
                except (KeyError, AttributeError) as e:
                    self._fontdict[c] = b'\0' * self._height
